from __future__ import annotations

from functools import cached_property

import numpy as np
import torch
from torch.linalg import eigvals

from qadence.blocks import PrimitiveBlock
from qadence.logger import get_logger

logger = get_logger(__name__)


class MatrixBlock(PrimitiveBlock):
    """
    Generates a MatrixBlock from a given matrix.

    Arguments:
        matrix (torch.Tensor | np.ndarray): The matrix from which to create the MatrixBlock.
        qubit_support (tuple[int]): The qubit_support of the block.

    Examples:
    ```python exec="on" source="material-block" result="json"
        import torch

    from qadence.circuit import QuantumCircuit
    from qadence.types import BackendName, DiffMode
    from qadence.blocks.matrix import MatrixBlock
    from qadence.models import QuantumModel
    from qadence.operations import X, Z
    from qadence.states import random_state

    n_qubits = 1
    XMAT = torch.tensor([[0, 1], [1, 0]], dtype=torch.cdouble)
    state = random_state(n_qubits)
    matblock = MatrixBlock(XMAT, (0,))

    qm_mat = QuantumModel(
        circuit=QuantumCircuit(n_qubits, matblock),
        observable=Z(0),
        backend=BackendName.PYQTORCH,
        diff_mode=DiffMode.AD,
    )
    qm = QuantumModel(
        circuit=QuantumCircuit(n_qubits, X(0)),
        observable=Z(0),
        backend=BackendName.PYQTORCH,
        diff_mode=DiffMode.AD,
    )
    wf_mat = qm_mat.run({}, state)
    exp_mat = qm_mat.expectation({})
    wf = qm.run({}, state)
    exp = qm.expectation({})

    assert torch.all(torch.isclose(wf_mat, wf)) and torch.isclose(exp, exp_mat)
    ```
    """

    name = "MatrixBlock"
    matrix: torch.Tensor

    def __init__(self, matrix: torch.Tensor | np.ndarray, qubit_support: tuple[int, ...]) -> None:
        if isinstance(matrix, np.ndarray):
            matrix = torch.tensor(matrix)
        if matrix.ndim == 3 and matrix.size(0) == 1:
            matrix = matrix.squeeze(0)
        if not matrix.ndim == 2:
            raise TypeError("Please provide a 2D matrix.")
        if not self.is_square(matrix):
            raise TypeError("Please provide a square matrix.")
        if not self.is_hermitian(matrix):
            logger.warning("Provided matrix is not hermitian.")
        if not self.is_unitary(matrix):
            logger.warning("Provided matrix is not unitary.")
        self.matrix = matrix.clone()
        super().__init__(qubit_support)

    @cached_property
    def eigenvalues_generator(self) -> torch.Tensor:
        return torch.log(self.eigenvalues) * 1j

    @property
    def eigenvalues(self) -> torch.Tensor:
        ev = eigvals(self.matrix)
        _, indices = torch.sort(ev.real)
        return ev[indices]

    @property
    def n_qubits(self) -> int:
        return np.log2(self.matrix.size()[0])  # type:ignore[no-any-return]

    @staticmethod
    def is_square(m: torch.Tensor) -> bool:
        return m.shape[0] == m.shape[1]  # type:ignore[no-any-return]

    @staticmethod
    def is_hermitian(m: torch.Tensor) -> bool:
        return MatrixBlock.is_square(m) and torch.allclose(
            m.t().conj(), m
        )  # type:ignore[no-any-return]

    @staticmethod
    def is_unitary(m: torch.Tensor) -> bool:
        if not MatrixBlock.is_square(m):
            return False
        prod = torch.mm(m, m.t().conj())
        i = torch.eye(m.shape[0], dtype=torch.complex128)
        return torch.allclose(prod, i)  # type:ignore[no-any-return]

    def expand_to(self, n_qubits: int = 1) -> torch.Tensor:
        from qadence.blocks.block_to_tensor import _fill_identities

        if n_qubits > 1:
            return _fill_identities(
                self.matrix, self.qubit_support, tuple([i for i in range(n_qubits)])
            )
        return self.matrix



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from jax import Array, device_get
from sympy import Expr
from sympy2jax import SymbolicModule as JaxSympyModule
from torch import Tensor, cdouble, from_numpy

from qadence.types import ParamDictType


def jarr_to_tensor(arr: Array, dtype: Any = cdouble) -> Tensor:
    return from_numpy(device_get(arr)).to(dtype=dtype)


def tensor_to_jnp(tensor: Tensor, dtype: Any = jnp.complex128) -> Array:
    return (
        jnp.array(tensor.numpy(), dtype=dtype)
        if not tensor.requires_grad
        else jnp.array(tensor.detach().numpy(), dtype=dtype)
    )


def values_to_jax(param_values: dict[str, Tensor]) -> dict[str, Array]:
    return {key: jnp.array(value.detach().numpy()) for key, value in param_values.items()}


def jaxify(expr: Expr) -> JaxSympyModule:
    return JaxSympyModule(expr)


def unhorqify(state: Array) -> Array:
    """Convert a state of shape [2] * n_qubits + [batch_size] to (batch_size, 2**n_qubits)."""
    return jnp.ravel(state)


def uniform_batchsize(param_values: ParamDictType) -> ParamDictType:
    max_batch_size = max(p.size for p in param_values.values())
    batched_values = {
        k: (v if v.size == max_batch_size else v.repeat(max_batch_size))
        for k, v in param_values.items()
    }
    return batched_values



from __future__ import annotations

from typing import Any, Union

from qadence.types import QubitSupportType


def _is_valid_support(t: Any) -> bool:
    return isinstance(t, tuple) and all(i >= 0 for i in t)


class QubitSupport(tuple):
    def __new__(cls, *support: Union[QubitSupportType, str, int, tuple]) -> QubitSupport:
        if len(support) == 1:
            if isinstance(support[0], tuple):
                return QubitSupport(*support[0])
            if support[0] == "global":
                support = (QubitSupportType.GLOBAL,)
                valid = True
            elif support[0] >= 0:  # type: ignore[operator]
                valid = True
            else:
                valid = False
        else:
            valid = _is_valid_support(support)

        if not valid:
            raise ValueError(
                "QubitSupport can be a tuple of ints or 'global'. For example:ℕn"
                "QubitSupport(1,2,3) or QubitSupport('global')\n"
                f"Found: {support}"
            )
        return super(QubitSupport, cls).__new__(cls, support)  # type: ignore[arg-type]

    def __add__(self, other: Any) -> QubitSupport:
        if not isinstance(other, tuple):
            raise ValueError(f"Cannot add type '{type(other)}' to QubitSupport.")
        if self == other:
            return self
        elif self == ("global",):
            return QubitSupport(*range(max(other) + 1)) if len(other) else QubitSupport("global")
        elif other == ("global",):
            return QubitSupport(*range(max(self) + 1)) if len(self) else QubitSupport("global")
        else:
            return QubitSupport(tuple({*self, *other}))

    def __radd__(self, other: Any) -> QubitSupport:
        return self.__add__(other)

    @property
    def is_global(self) -> bool:
        return self == ("global",)

    def is_disjoint(self, other: Any) -> bool:
        oth = QubitSupport(other)
        if self.is_global or oth.is_global:
            return False
        else:
            selfsup = set(self)
            othersup = set(oth)
            return selfsup.isdisjoint(othersup)



from __future__ import annotations

import sympy
import torch
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.tree import Tree
from torch import Tensor, cdouble

from qadence.blocks import (
    ControlBlock,
    ParametricControlBlock,
)
from qadence.blocks.utils import (
    add,  # noqa
    chain,
    kron,
)
from qadence.logger import get_logger
from qadence.parameters import (
    Parameter,
    evaluate,
)
from qadence.types import OpName, TNumber, TParameter

from .parametric import PHASE, RX, RY, RZ
from .primitive import SWAP, I, N, X, Y, Z

logger = get_logger(__name__)


class CNOT(ControlBlock):
    """The CNot, or CX, gate."""

    name = OpName.CNOT

    def __init__(self, control: int, target: int) -> None:
        self.generator = kron(N(control), X(target) - I(target))
        super().__init__((control,), X(target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0, 0, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1, 1, 1], dtype=cdouble)

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree


class MCZ(ControlBlock):
    name = OpName.MCZ

    def __init__(self, control: tuple[int, ...], target: int) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Z(target) - I(target))
        super().__init__(control, Z(target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat((torch.tensor(-2, dtype=cdouble), torch.zeros(2**self.n_qubits - 1)))

    @property
    def eigenvalues(self) -> Tensor:
        torch.cat((torch.tensor(-1, dtype=cdouble), torch.ones(2**self.n_qubits - 1)))

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree


class CZ(MCZ):
    """The CZ gate."""

    name = OpName.CZ

    def __init__(self, control: int, target: int) -> None:
        super().__init__((control,), target)


class MCRX(ParametricControlBlock):
    name = OpName.MCRX

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], X(target))
        super().__init__(control, RX(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRX(MCRX):
    """The CRX gate."""

    name = OpName.CRX

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class MCRY(ParametricControlBlock):
    name = OpName.MCRY

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Y(target))
        super().__init__(control, RY(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRY(MCRY):
    """The CRY gate."""

    name = OpName.CRY

    def __init__(
        self,
        control: int,
        target: int,
        parameter: TParameter,
    ):
        super().__init__((control,), target, parameter)


class MCRZ(ParametricControlBlock):
    name = OpName.MCRZ

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Z(target))
        super().__init__(control, RZ(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.zeros(2**self.n_qubits - 2), torch.tensor([1, -1], dtype=cdouble))
        )

    @property
    def eigenvalues(self) -> Tensor:
        val = evaluate(self.parameters.parameter, as_torch=True)
        lmbd = torch.cos(val / 2.0) - 1j * torch.sin(val / 2.0)
        return torch.cat((torch.ones(2**self.n_qubits - 2), lmbd, lmbd.conj()))


class CRZ(MCRZ):
    """The CRZ gate."""

    name = OpName.CRZ

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class MCPHASE(ParametricControlBlock):
    name = OpName.MCPHASE

    def __init__(
        self,
        control: tuple[int, ...],
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], Z(target) - I(target))
        super().__init__(control, PHASE(target, parameter))

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.cat(
            (torch.tensor([-2, 0], dtype=cdouble), (torch.zeros(2**self.n_qubits - 2)))
        )

    @property
    def eigenvalues(self) -> Tensor:
        v = evaluate(self.parameters.parameter, as_torch=True)
        return torch.cat((torch.ones(2**self.n_qubits - 1), torch.exp(1j * v)))

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree

    def __ascii__(self, console: Console) -> RenderableType:
        (target, control) = self.qubit_support
        h = abs(target - control) + 1
        return Panel(self._block_title, expand=False, height=3 * h)


class CPHASE(MCPHASE):
    """The CPHASE gate."""

    name = OpName.CPHASE

    def __init__(
        self,
        control: int,
        target: int,
        parameter: Parameter | TNumber | sympy.Expr | str,
    ):
        super().__init__((control,), target, parameter)


class CSWAP(ControlBlock):
    """The CSWAP (Control-SWAP) gate."""

    name = OpName.CSWAP

    def __init__(self, control: int | tuple[int, ...], target1: int, target2: int) -> None:
        if isinstance(control, tuple):
            control = control[0]

        a00m = -N(target=control)
        a00p = -N(target=control, state="0")
        a11 = -N(target=target1)
        a22 = -N(target=target2, state="0")
        a12 = 0.5 * (chain(X(target1), Z(target1)) + X(target1))
        a21 = 0.5 * (chain(Z(target2), X(target2)) + X(target2))
        no_effect = kron(a00m, I(target1), I(target2))
        swap_effect = (
            kron(a00p, -1.0 * a22, a11)
            + kron(a00p, -1.0 * a11, a22)
            + kron(a00p, a12, a21)
            + kron(a00p, a21, a12)
        )
        self.generator = no_effect + swap_effect
        super().__init__((control,), SWAP(target1, target2))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor((1, -1, 1, 1, 1, 1, 1, 1), dtype=torch.cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor((1, -1, 1, 1, 1, 1, 1, 1), dtype=torch.cdouble)

    @property
    def nqubits(self) -> int:
        return 3


class Toffoli(ControlBlock):
    name = OpName.TOFFOLI

    def __init__(self, control: tuple[int, ...], target: int) -> None:
        self.generator = kron(*[N(qubit) for qubit in control], X(target) - I(target))
        super().__init__(control, X(target))

    @property
    def n_qubits(self) -> int:
        return len(self.qubit_support)

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor(
            [-2, *[0 for _ in range(2 ** len(self.qubit_support) - 1)]], dtype=cdouble
        )

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor(
            [-1, *[1 for _ in range(2 ** len(self.qubit_support) - 1)]], dtype=cdouble
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Tuple

import numpy as np
import sympy
import torch
from sympy import Basic
from torch import cdouble

from qadence.blocks.analog import (
    AnalogBlock,
    ConstantAnalogRotation,
    InteractionBlock,
    QubitSupport,
)
from qadence.blocks.utils import (
    add,  # noqa
    kron,
)
from qadence.logger import get_logger
from qadence.parameters import (
    Parameter,
    ParamMap,
)
from qadence.types import PI, OpName, TNumber, TParameter

from .ham_evo import HamEvo
from .primitive import I, X, Z

logger = get_logger(__name__)


class AnalogSWAP(HamEvo):
    """
    Single time-independent Hamiltonian evolution over a Rydberg Ising.

    hamiltonian yielding a SWAP (up to global phase).

    Derived from
    [Bapat et al.](https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.4.L012023)
    where it is applied to XX-type Hamiltonian
    """

    name = OpName.ANALOGSWAP

    def __init__(self, control: int, target: int, parameter: TParameter = 3 * PI / 4):
        rydberg_ising_hamiltonian_generator = (
            4.0 * kron((I(control) - Z(control)) / 2.0, (I(target) - Z(target)) / 2.0)
            + (2.0 / 3.0) * np.sqrt(2.0) * X(control)
            + (2.0 / 3.0) * np.sqrt(2.0) * X(target)
            + (1.0 + np.sqrt(5.0) / 3) * Z(control)
            + (1.0 + np.sqrt(5.0) / 3) * Z(target)
        )
        super().__init__(rydberg_ising_hamiltonian_generator, parameter, (control, target))


def _cast(T: Any, val: Any) -> Any:
    return val if isinstance(val, T) else T(val)


def AnalogInteraction(
    duration: TNumber | sympy.Basic,
    qubit_support: str | QubitSupport | tuple = "global",
    add_pattern: bool = True,
) -> InteractionBlock:
    """Evolution of the interaction term for a register of qubits.

    Constructs a [`InteractionBlock`][qadence.blocks.analog.InteractionBlock].

    Arguments:
        duration: Time to evolve the interaction for in nanoseconds.
        qubit_support: Qubits the `InteractionBlock` is applied to. Can be either
            `"global"` to evolve the interaction block to all qubits or a tuple of integers.
        add_pattern: False disables the semi-local addressing pattern
            for the execution of this specific block.

    Returns:
        a `InteractionBlock`
    """
    q = _cast(QubitSupport, qubit_support)
    ps = ParamMap(duration=duration)
    return InteractionBlock(parameters=ps, qubit_support=q, add_pattern=add_pattern)


def wait(
    duration: TNumber | sympy.Basic,
    qubit_support: str | QubitSupport | tuple = "global",
    add_pattern: bool = True,
) -> InteractionBlock:
    logger.warning("The alias `wait` is deprecated, please use `AnalogInteraction`")
    return AnalogInteraction(duration, qubit_support, add_pattern)


# FIXME: clarify the usage of this gate, rename more formally, and implement in PyQ
@dataclass(eq=False, repr=False)
class AnalogEntanglement(AnalogBlock):
    parameters: ParamMap = ParamMap(duration=1.0)
    qubit_support: QubitSupport = QubitSupport("global")

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        return torch.tensor([0.0], dtype=cdouble)

    @property
    def duration(self) -> Basic:
        return self.parameters.duration


def entangle(
    duration: Any,
    qubit_support: str | QubitSupport | Tuple = "global",
) -> AnalogEntanglement:
    q = _cast(QubitSupport, qubit_support)
    ps = ParamMap(duration=duration)
    return AnalogEntanglement(parameters=ps, qubit_support=q)


def AnalogRot(
    duration: float | str | Parameter,
    omega: float | str | Parameter = 0,
    delta: float | str | Parameter = 0,
    phase: float | str | Parameter = 0,
    qubit_support: str | QubitSupport | Tuple = "global",
    add_pattern: bool = True,
) -> ConstantAnalogRotation:
    """General analog rotation operation.

    Arguments:
        duration: Duration of the rotation [ns].
        omega: Rotation frequency [rad/μs]
        delta: Rotation frequency [rad/μs]
        phase: Phase angle [rad]
        qubit_support: Defines the (local/global) qubit support
        add_pattern: False disables the semi-local addressing pattern
            for the execution of this specific block.

    Returns:
        ConstantAnalogRotation
    """

    if omega == 0 and delta == 0:
        raise ValueError("Parameters omega and delta cannot both be 0.")

    q = _cast(QubitSupport, qubit_support)
    duration = Parameter(duration)
    omega = Parameter(omega)
    delta = Parameter(delta)
    phase = Parameter(phase)
    h_norm = sympy.sqrt(omega**2 + delta**2)
    alpha = duration * h_norm / 1000
    ps = ParamMap(
        alpha=alpha, duration=duration, omega=omega, delta=delta, phase=phase, h_norm=h_norm
    )
    return ConstantAnalogRotation(parameters=ps, qubit_support=q, add_pattern=add_pattern)


def _analog_rot(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple,
    phase: float,
    add_pattern: bool = True,
) -> ConstantAnalogRotation:
    q = _cast(QubitSupport, qubit_support)
    # assuming some arbitrary omega = π rad/μs
    alpha = _cast(Parameter, angle)
    delta = 0
    omega = PI
    duration = alpha / omega * 1000
    h_norm = sympy.sqrt(omega**2 + delta**2)

    # FIXME: once https://github.com/pasqal-io/qadence/issues/150 is fixed set default duration
    # in the function arguments to:
    # duration = Parameter(160)
    # and compute omega like this:
    # omega = alpha / duration * 1000
    ps = ParamMap(
        alpha=alpha, duration=duration, omega=omega, delta=delta, phase=phase, h_norm=h_norm
    )
    return ConstantAnalogRotation(parameters=ps, qubit_support=q, add_pattern=add_pattern)


def AnalogRX(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple = "global",
    add_pattern: bool = True,
) -> ConstantAnalogRotation:
    """Analog X rotation.

    Shorthand for [`AnalogRot`][qadence.operations.AnalogRot]:

    ```python
    φ=2.4; Ω=π; t = φ/Ω * 1000
    AnalogRot(duration=t, omega=Ω)
    ```

    Arguments:
        angle: Rotation angle [rad]
        qubit_support: Defines the (local/global) qubit support

    Returns:
        ConstantAnalogRotation
    """
    return _analog_rot(angle, qubit_support, phase=0, add_pattern=add_pattern)


def AnalogRY(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple = "global",
    add_pattern: bool = True,
) -> ConstantAnalogRotation:
    """Analog Y rotation.

    Shorthand for [`AnalogRot`][qadence.operations.AnalogRot]:

    ```python
    φ=2.4; Ω=π; t = φ/Ω * 1000
    AnalogRot(duration=t, omega=Ω, phase=-π/2)
    ```
    Arguments:
        angle: Rotation angle [rad]
        qubit_support: Defines the (local/global) qubit support

    Returns:
        ConstantAnalogRotation
    """
    return _analog_rot(angle, qubit_support, phase=-PI / 2, add_pattern=add_pattern)


def AnalogRZ(
    angle: float | str | Parameter,
    qubit_support: str | QubitSupport | Tuple = "global",
    add_pattern: bool = True,
) -> ConstantAnalogRotation:
    """Analog Z rotation. Shorthand for [`AnalogRot`][qadence.operations.AnalogRot]:
    ```
    φ=2.4; δ=π; t = φ/δ * 100)
    AnalogRot(duration=t, delta=δ, phase=π/2)
    ```
    """
    q = _cast(QubitSupport, qubit_support)
    alpha = _cast(Parameter, angle)
    delta = PI
    omega = 0
    duration = alpha / delta * 1000
    h_norm = sympy.sqrt(omega**2 + delta**2)
    ps = ParamMap(
        alpha=alpha, duration=duration, omega=omega, delta=delta, phase=0.0, h_norm=h_norm
    )
    return ConstantAnalogRotation(qubit_support=q, parameters=ps, add_pattern=add_pattern)



from __future__ import annotations

from copy import deepcopy
from functools import cached_property
from typing import Any, Union

import numpy as np
import sympy
import torch
from torch import Tensor

from qadence.blocks import (
    AbstractBlock,
    TimeEvolutionBlock,
)
from qadence.blocks.block_to_tensor import block_to_tensor
from qadence.blocks.utils import (
    add,  # noqa
    block_is_commuting_hamiltonian,
    block_is_qubit_hamiltonian,
    chain,
    expressions,
)
from qadence.decompose import lie_trotter_suzuki
from qadence.logger import get_logger
from qadence.parameters import (
    Parameter,
    ParamMap,
    evaluate,
    extract_original_param_entry,
)
from qadence.types import LTSOrder, OpName, TGenerator, TParameter
from qadence.utils import eigenvalues

logger = get_logger(__name__)


class HamEvo(TimeEvolutionBlock):
    """
    A block implementing the Hamiltonian evolution operation H where:

        H = exp(-iG, t)
    where G represents a square generator and t represents the time parameter
    which can be parametrized.

    Arguments:
        generator: Either a AbstractBlock, torch.Tensor or numpy.ndarray.
        parameter: A scalar or vector of numeric or torch.Tensor type.
        qubit_support: The qubits on which the evolution will be performed on.

    Examples:

    ```python exec="on" source="material-block" result="json"
    from qadence import RX, HamEvo, run, PI
    import torch
    hevo = HamEvo(generator=RX(0, PI), parameter=torch.rand(2))
    print(run(hevo))
    # Now lets use a torch.Tensor as a generator, Now we have to pass the support
    gen = torch.rand(2,2, dtype=torch.complex128)
    hevo = HamEvo(generator=gen, parameter=torch.rand(2), qubit_support=(0,))
    print(run(hevo))
    ```
    """

    name = OpName.HAMEVO
    draw_generator: bool = False

    def __init__(
        self,
        generator: Union[TGenerator, AbstractBlock],
        parameter: TParameter,
        qubit_support: tuple[int, ...] = None,
    ):
        gen_exprs = {}
        if qubit_support is None and not isinstance(generator, AbstractBlock):
            raise ValueError("You have to supply a qubit support for non-block generators.")
        super().__init__(qubit_support if qubit_support else generator.qubit_support)
        if isinstance(generator, AbstractBlock):
            qubit_support = generator.qubit_support
            if generator.is_parametric:
                gen_exprs = {str(e): e for e in expressions(generator)}
        elif isinstance(generator, torch.Tensor):
            msg = "Please provide a square generator."
            if len(generator.shape) == 2:
                assert generator.shape[0] == generator.shape[1], msg
            elif len(generator.shape) == 3:
                assert generator.shape[1] == generator.shape[2], msg
                assert generator.shape[0] == 1, "Qadence doesnt support batched generators."
            else:
                raise TypeError(
                    "Only 2D or 3D generators are supported.\
                                In case of a 3D generator, the batch dim\
                                is expected to be at dim 0."
                )
            gen_exprs = {str(generator.__hash__()): generator}
        elif isinstance(generator, (sympy.Basic, sympy.Array)):
            gen_exprs = {str(generator): generator}
        else:
            raise TypeError(
                f"Generator of type {type(generator)} not supported.\
                            If you're using a numpy.ndarray, please cast it to a torch tensor."
            )
        ps = {"parameter": Parameter(parameter), **gen_exprs}
        self.parameters = ParamMap(**ps)
        self.generator = generator

    @classmethod
    def num_parameters(cls) -> int:
        return 2

    @cached_property
    def eigenvalues_generator(
        self, max_num_evals: int | None = None, max_num_gaps: int | None = None
    ) -> Tensor:
        if isinstance(self.generator, AbstractBlock):
            generator_tensor = block_to_tensor(self.generator)
        elif isinstance(self.generator, Tensor):
            generator_tensor = self.generator
        return eigenvalues(generator_tensor, max_num_evals, max_num_gaps)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.exp(
            -1j * evaluate(self.parameters.parameter, as_torch=True) * self.eigenvalues_generator
        )

    @property
    def n_qubits(self) -> int:
        if isinstance(self.generator, Tensor):
            n_qubits = int(np.log2(self.generator.shape[-1]))
        else:
            n_qubits = self.generator.n_qubits  # type: ignore [union-attr]

        return n_qubits

    def dagger(self) -> Any:
        p = list(self.parameters.expressions())[0]
        return HamEvo(deepcopy(self.generator), -extract_original_param_entry(p))

    def digital_decomposition(self, approximation: LTSOrder = LTSOrder.ST4) -> AbstractBlock:
        """Decompose the Hamiltonian evolution into digital gates.

        Args:
            approximation (str, optional): Choose the type of decomposition. Defaults to "st4".
                Available types are:
                * 'basic' = apply first-order Trotter formula and decompose each term of
                    the exponential into digital gates. It is exact only if applied to an
                    operator whose terms are mutually commuting.
                * 'st2' = Trotter-Suzuki 2nd order formula for approximating non-commuting
                    Hamiltonians.
                * 'st4' = Trotter-Suzuki 4th order formula for approximating non-commuting
                    Hamiltonians.

        Returns:
            AbstractBlock: a block with the digital decomposition
        """

        # psi(t) = exp(-i * H * t * psi0)
        # psi(t) = exp(-i * lambda * t * psi0)
        # H = sum(Paulin) + sum(Pauli1*Pauli2)
        logger.info("Quantum simulation of the time-independent Schrödinger equation.")

        blocks = []

        # how to change the type/dict to enum effectively

        # when there is a term including non-commuting matrices use st2 or st4

        # 1) should check that the given generator respects the constraints
        # single-qubit gates

        assert isinstance(
            self.generator, AbstractBlock
        ), "Only a generator represented as a block can be decomposed"

        if block_is_qubit_hamiltonian(self.generator):
            try:
                block_is_commuting_hamiltonian(self.generator)
                approximation = LTSOrder.BASIC  # use the simpler approach if the H is commuting
            except TypeError:
                logger.warning(
                    """Non-commuting terms in the Pauli operator.
                    The Suzuki-Trotter approximation is applied."""
                )

            blocks.extend(
                lie_trotter_suzuki(
                    block=self.generator,
                    parameter=self.parameters.parameter,
                    order=LTSOrder[approximation],
                )
            )

            # 2) return an AbstractBlock instance with the set of gates
            # resulting from the decomposition

            return chain(*blocks)
        else:
            raise NotImplementedError(
                "The current digital decomposition can be applied only to Pauli Hamiltonians."
            )



from __future__ import annotations

from copy import deepcopy
from functools import reduce, singledispatch
from typing import Generator, Type, overload

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    CompositeBlock,
    KronBlock,
    ScaleBlock,
)
from qadence.blocks.utils import _construct
from qadence.circuit import QuantumCircuit


def _flat_blocks(block: AbstractBlock, T: Type) -> Generator:
    """Constructs a generator that flattens nested `CompositeBlock`s of type `T`.

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence.transpile.block import _flat_blocks
    from qadence.blocks import ChainBlock
    from qadence import chain, X

    x = chain(chain(chain(X(0)), X(0)))
    assert tuple(_flat_blocks(x, ChainBlock)) == (X(0), X(0))
    ```
    """
    if isinstance(block, T):
        # here we do the flattening
        for b in block.blocks:
            if isinstance(b, T):
                yield from _flat_blocks(b, T)
            else:
                yield flatten(b, [T])
    elif isinstance(block, CompositeBlock):
        # here we make sure that we don't get stuck at e.g. `KronBlock`s if we
        # want to flatten `ChainBlock`s
        yield from (flatten(b, [T]) for b in block.blocks)
    elif isinstance(block, ScaleBlock):
        blk = deepcopy(block)
        blk.block = flatten(block.block, [T])
        yield blk
    else:
        yield block


@overload
def flatten(
    circuit: QuantumCircuit, types: list = [ChainBlock, KronBlock, AddBlock]
) -> QuantumCircuit:
    ...


@overload
def flatten(block: AbstractBlock, types: list = [ChainBlock, KronBlock, AddBlock]) -> AbstractBlock:
    ...


@singledispatch
def flatten(
    circ_or_block: AbstractBlock | QuantumCircuit, types: list = [ChainBlock, KronBlock, AddBlock]
) -> AbstractBlock | QuantumCircuit:
    raise NotImplementedError(f"digitalize is not implemented for {type(circ_or_block)}")


@flatten.register  # type: ignore[attr-defined]
def _(block: AbstractBlock, types: list = [ChainBlock, KronBlock, AddBlock]) -> AbstractBlock:
    """Flattens the given types of `CompositeBlock`s if possible.

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import chain, kron, X
    from qadence.transpile import flatten
    from qadence.blocks import ChainBlock, KronBlock, AddBlock

    x = chain(chain(chain(X(0))), kron(kron(X(0))))

    # flatten only `ChainBlock`s
    assert flatten(x, [ChainBlock]) == chain(X(0), kron(kron(X(0))))

    # flatten `ChainBlock`s and `KronBlock`s
    assert flatten(x, [ChainBlock, KronBlock]) == chain(X(0), kron(X(0)))

    # flatten `AddBlock`s (does nothing in this case)
    assert flatten(x, [AddBlock]) == x
    ```
    """
    if isinstance(block, CompositeBlock):

        def fn(b: AbstractBlock, T: Type) -> AbstractBlock:
            return _construct(type(block), tuple(_flat_blocks(b, T)))

        return reduce(fn, types, block)  # type: ignore[arg-type]
    elif isinstance(block, ScaleBlock):
        blk = deepcopy(block)
        blk.block = flatten(block.block, types=types)
        return blk
    else:
        return block


@flatten.register  # type: ignore[attr-defined]
def _(circuit: QuantumCircuit, types: list = [ChainBlock, KronBlock, AddBlock]) -> QuantumCircuit:
    return QuantumCircuit(circuit.n_qubits, flatten(circuit.block, types=types))



from __future__ import annotations

from collections import Counter
from enum import Enum

import torch
from torch import Tensor
from torch.distributions import normal, poisson, uniform

from qadence.logger import get_logger

logger = get_logger(__name__)


class WhiteNoise(Enum):
    """White noise distributions."""

    UNIFORM = staticmethod(uniform.Uniform(low=0.0, high=1.0))
    """Uniform white noise."""

    GAUSSIAN = staticmethod(normal.Normal(loc=0.0, scale=1.0))
    """Gaussian white noise."""

    POISSON = staticmethod(poisson.Poisson(rate=0.1))
    """Poisson white noise."""


def bitstring_to_tensor(bitstring: str) -> Tensor:
    """
    A helper function to convert bit strings to torch.Tensor.

    Args:
        bitstring:  A str format of a bit string.

    Returns:
        A torch.Tensor out of the input bit string.
    """
    return torch.as_tensor(list(map(int, bitstring)))


def tensor_to_bitstring(bitstring: Tensor) -> str:
    """
    A helper function to convert torch.Tensor to bit strings.

    Args:
        bitstring: A torch.Tensor format of a bit string.

    Returns:
        A str out of the input bit string.
    """
    return "".join(list(map(str, bitstring.detach().tolist())))


def bit_flip(bit: Tensor, cond: Tensor) -> Tensor:
    """
    A helper function that reverses the states 0 and 1 in the bit string.

    Args:
        bit: A integer-value bit in a bitstring to be inverted.
        cond: A Bool value of whether or not a bit should be flipped.

    Returns:
        The inverse value of the input bit
    """
    return torch.where(
        cond,
        torch.where(
            bit == torch.zeros(1, dtype=torch.int64),
            torch.ones(1, dtype=torch.int64),
            torch.zeros(1, dtype=torch.int64),
        ),
        bit,
    )


def sample_to_matrix(sample: dict) -> Tensor:
    """
    A helper function that maps a sample dict to a bit string array.

    Args:
        sample: A dictionary with bit stings as keys and values
        as their counts.

    Returns: A torch.Tensor of bit strings n_shots x n_qubits.
    """

    return torch.concatenate(
        list(
            map(
                lambda bitstring: torch.broadcast_to(
                    bitstring_to_tensor(bitstring), [sample[bitstring], len(bitstring)]
                ),
                sample.keys(),
            )
        )
    )


def create_noise_matrix(
    noise_distribution: torch.distributions, n_shots: int, n_qubits: int
) -> Tensor:
    """
    A helper function that creates a noise matrix for bit string corruption.

    NB: The noise matrix is not square, as all bits are considered independent.

    Args:
        noise_distribution: Torch statistical distribution one of Gaussian,
        Uniform, or Poisson.
        n_shots: Number of shots/samples.
        n_qubits: Number of qubits

    Returns:
        A sample out of the requested distribution given the number of shots/samples.
    """
    # the noise_matrix should be available to the user if they want to do error correction
    return noise_distribution.sample([n_shots, n_qubits])


def bs_corruption(
    err_idx: Tensor,
    sample: Tensor,
) -> Counter:
    """
    A function that incorporates the expected readout error in a sample of bit strings.

    given a noise matrix.

    Args:
        err_idx: A Boolean array of bit string indices that need to be corrupted.
        sample: A torch.Tensor of bit strings n_shots x n_qubits.

    Returns:
        A counter of bit strings after readout corruption.
    """

    func = torch.func.vmap(bit_flip)

    return Counter([tensor_to_bitstring(k) for k in func(sample, err_idx)])


def create_confusion_matrices(noise_matrix: Tensor, error_probability: float) -> Tensor:
    confusion_matrices = []
    for i in range(noise_matrix.size()[1]):
        column_tensor = noise_matrix[:, i]
        flip_proba = column_tensor[column_tensor < error_probability].mean().item()
        confusion_matrix = torch.tensor(
            [[1.0 - flip_proba, flip_proba], [flip_proba, 1.0 - flip_proba]], dtype=torch.float64
        )
        confusion_matrices.append(confusion_matrix)
    return torch.stack(confusion_matrices)


def add_noise(
    counters: list[Counter],
    n_qubits: int,
    n_shots: int = 1000,
    options: dict = dict(),
) -> list[Counter]:
    """
    Implements a simple uniform readout error model for position-independent bit string.

    corruption.

    Args:
        counters (list): Samples of bit string as Counters.
        n_qubits: Number of qubits in the bit string.
        n_shots: Number of shots to sample.
        options: A dict of options:
          seed: Random seed value if any.
          error_probability: Uniform error probability of wrong readout at any position
            in the bit strings.
          noise_distribution: Noise distribution.
          noise_matrix: An input noise matrix if known.

    Returns:
        Samples of corrupted bit strings as list[Counter].
    """

    seed = options.get("seed", None)
    error_probability = options.get("error_probability")
    noise_distribution = options.get("noise_distribution", WhiteNoise.UNIFORM)
    noise_matrix = options.get("noise_matrix")

    # option for reproducibility
    if seed is not None:
        torch.manual_seed(seed)

    if error_probability is None:
        error_probability = 0.1
        # Return the default error probability for mitigation purposes.
        options["default_error_probability"] = error_probability
    if noise_matrix is None:
        # assumes that all bits can be flipped independently of each other
        noise_matrix = create_noise_matrix(noise_distribution, n_shots, n_qubits)
        confusion_matrices = create_confusion_matrices(
            noise_matrix=noise_matrix, error_probability=error_probability
        )
        # Return the generated noise matrix for mitigation purposes.
        options["confusion_matrices"] = confusion_matrices
    else:
        # check noise_matrix shape and values
        assert (
            noise_matrix.shape[0] == noise_matrix.shape[1]
        ), "The error probabilities matrix needs to be square."
        assert noise_matrix.shape == (
            n_qubits,
            n_qubits,
        ), "The error probabilities matrix needs to be n_qubits x n_qubits."

    # the simplest approach - an event occurs if its probability is higher than expected
    # by random chance
    err_idx = torch.as_tensor(noise_matrix < error_probability)

    corrupted_bitstrings = []
    for counter in counters:
        sample = sample_to_matrix(counter)
        corrupted_bitstrings.append(bs_corruption(err_idx=err_idx, sample=sample))
    return corrupted_bitstrings



from __future__ import annotations

from typing import Union

import numpy as np
import torch
from rich.console import Console
from rich.padding import Padding
from torch import Tensor, cdouble, tensor

from qadence.blocks import (
    AbstractBlock,
    PrimitiveBlock,
)
from qadence.blocks.primitive import ProjectorBlock
from qadence.blocks.utils import (
    chain,
    kron,
)
from qadence.logger import get_logger
from qadence.parameters import (
    Parameter,
)
from qadence.types import OpName, TNumber

logger = get_logger(__name__)


class X(PrimitiveBlock):
    """The X gate."""

    name = OpName.X

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)


class Y(PrimitiveBlock):
    """The Y gate."""

    name = OpName.Y

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)


class Z(PrimitiveBlock):
    """The Z gate."""

    name = OpName.Z

    def __init__(self, target: int):
        super().__init__((target,))

    @property
    def generator(self) -> AbstractBlock:
        return self

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([-1, 1], dtype=cdouble)


class I(PrimitiveBlock):
    """The identity gate."""

    name = OpName.I

    def __init__(self, target: int):
        super().__init__((target,))

    def __ixor__(self, other: AbstractBlock | int) -> AbstractBlock:
        if not isinstance(other, AbstractBlock):
            raise ValueError(
                f"Can only initialize a kron block with another block. Got {type(other)}."
            )
        return other

    def __imul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        if not isinstance(other, AbstractBlock):
            raise ValueError(
                "In-place multiplication is available " "only for AbstractBlock instances"
            )
        return other

    @property
    def generator(self) -> AbstractBlock:
        return I(*self.qubit_support)

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.ones(2, dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.ones(2, dtype=cdouble)

    def __ascii__(self, console: Console) -> Padding:
        return Padding("──────", (1, 1, 1, 1))


class Projector(ProjectorBlock):
    """The projector operator."""

    name = OpName.PROJ

    def __init__(
        self,
        ket: str,
        bra: str,
        qubit_support: int | tuple[int, ...],
    ):
        super().__init__(ket=ket, bra=bra, qubit_support=qubit_support)

    @property
    def generator(self) -> None:
        raise ValueError("Property `generator` not available for non-unitary operator.")

    @property
    def eigenvalues_generator(self) -> None:
        raise ValueError("Property `eigenvalues_generator` not available for non-unitary operator.")


class N(Projector):
    """The N = (1/2)(I-Z) operator."""

    name = OpName.N

    def __init__(self, target: int, state: str = "1"):
        super().__init__(ket=state, bra=state, qubit_support=(target,))

    @property
    def generator(self) -> None:
        raise ValueError("Property `generator` not available for non-unitary operator.")

    @property
    def eigenvalues_generator(self) -> None:
        raise ValueError("Property `eigenvalues_generator` not available for non-unitary operator.")

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)


class S(PrimitiveBlock):
    """The S / Phase gate."""

    name = OpName.S

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([1, 1j], dtype=cdouble)

    def dagger(self) -> SDagger:
        return SDagger(*self.qubit_support)


class SDagger(PrimitiveBlock):
    """The Hermitian adjoint/conjugate transpose of the S / Phase gate."""

    name = OpName.SDAGGER

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return tensor([1, -1j], dtype=cdouble)

    def dagger(self) -> S:
        return S(*self.qubit_support)


class H(PrimitiveBlock):
    """The Hadamard or H gate."""

    name = OpName.H

    def __init__(self, target: int):
        self.generator = (1 / np.sqrt(2)) * (X(target) + Z(target) - np.sqrt(2) * I(target))
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1], dtype=cdouble)


class Zero(PrimitiveBlock):
    name = OpName.ZERO

    def __init__(self) -> None:
        self.generator = 0 * I(0)
        super().__init__((0,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.zeros(2, dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.zeros(2, dtype=cdouble)

    def __add__(self, other: AbstractBlock) -> AbstractBlock:
        return other

    def __iadd__(self, other: AbstractBlock) -> AbstractBlock:
        return other

    def __sub__(self, other: AbstractBlock) -> AbstractBlock:
        return -other

    def __isub__(self, other: AbstractBlock) -> AbstractBlock:
        return -other

    def __mul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __imul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __ixor__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self

    def __pow__(self, power: int) -> AbstractBlock:
        return self


class T(PrimitiveBlock):
    """The T gate."""

    name = OpName.T

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([1.0, torch.sqrt(torch.tensor([1j]))], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 1

    def dagger(self) -> TDagger:
        return TDagger(*self.qubit_support)


class TDagger(PrimitiveBlock):
    """The Hermitian adjoint/conjugate transpose of the T gate."""

    # FIXME: this gate is not support by any backend
    name = "T_dagger"

    def __init__(self, target: int):
        self.generator = I(target) - Z(target)
        super().__init__((target,))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([0, 1], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([1.0, torch.sqrt(torch.tensor([-1j]))], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 1

    def dagger(self) -> T:
        return T(*self.qubit_support)


class SWAP(PrimitiveBlock):
    """The SWAP gate."""

    name = OpName.SWAP

    def __init__(self, control: int, target: int) -> None:
        a11 = 0.5 * (Z(control) - I(control))
        a22 = -0.5 * (Z(target) + I(target))
        a12 = 0.5 * (chain(X(control), Z(control)) + X(control))
        a21 = 0.5 * (chain(Z(target), X(target)) + X(target))
        self.generator = (
            kron(-1.0 * a22, a11) + kron(-1.0 * a11, a22) + kron(a12, a21) + kron(a21, a12)
        )
        super().__init__((control, target))

    @property
    def eigenvalues_generator(self) -> Tensor:
        return torch.tensor([-2, 0, 0, 0], dtype=cdouble)

    @property
    def eigenvalues(self) -> Tensor:
        return torch.tensor([-1, 1, 1, 1], dtype=cdouble)

    @property
    def n_qubits(self) -> int:
        return 2

    @property
    def _block_title(self) -> str:
        c, t = self.qubit_support
        s = f"{self.name}({c}, {t})"
        return s if self.tag is None else (s + rf" \[tag: {self.tag}]")


TPauliBlock = Union[X, Y, Z, I, N]



from __future__ import annotations

import numpy as np
from pulser.parametrized.decorators import parametrize
from pulser.waveforms import ConstantWaveform

# determined by hardware team as a safe resolution
MAX_AMPLITUDE_SCALING = 0.1
EPS = 1e-9


class SquareWaveform(ConstantWaveform):
    def __init__(self, duration: int, value: float):
        super().__init__(duration, value)

    @classmethod
    @parametrize
    def from_area(
        cls,
        area: float,
        max_amp: float,
        duration_steps: int = 1,
        min_duration: int = 1,
    ) -> SquareWaveform:
        amp_steps = MAX_AMPLITUDE_SCALING * max_amp

        duration = max(
            duration_steps * np.round(area / (duration_steps * max_amp) * 1e3),
            min_duration,
        )
        amplitude = min(
            amp_steps * np.ceil(area / (amp_steps * duration) * 1e3),
            max_amp,
        )
        delta = np.abs(1e-3 * duration * amplitude - area)

        new_duration = duration + duration_steps
        new_amplitude = max(
            amp_steps * np.ceil(area / (amp_steps * new_duration) * 1e3),
            max_amp,
        )
        new_delta = np.abs(1e-3 * new_duration * new_amplitude - area)

        while new_delta < delta:
            duration = new_duration
            amplitude = new_amplitude
            delta = new_delta

            new_duration = duration + duration_steps
            new_amplitude = max(
                amp_steps * np.ceil(area / (amp_steps * new_duration) * 1e3),
                max_amp,
            )
            new_delta = np.abs(1e-3 * new_duration * new_amplitude - area)

        return cls(duration, amplitude)

    @classmethod
    @parametrize
    def from_duration(
        cls,
        duration: int,
        max_amp: float,
        duration_steps: int = 1,
        min_duration: int = 1,
    ) -> SquareWaveform:
        amp_steps = MAX_AMPLITUDE_SCALING * max_amp

        duration = max(
            duration_steps * np.round(duration / duration_steps),
            min_duration,
        )
        amplitude = min(
            amp_steps * np.ceil(max_amp / (amp_steps + EPS) * 1e3),
            max_amp,
        )

        return cls(duration, amplitude)



from __future__ import annotations

from typing import Any, Type, Union

import torch

from qadence.blocks import AbstractBlock, KronBlock, block_is_qubit_hamiltonian, chain, kron, tag
from qadence.constructors.hamiltonians import hamiltonian_factory
from qadence.operations import CNOT, CPHASE, CRX, CRY, CRZ, CZ, RX, RY, HamEvo
from qadence.parameters import Parameter
from qadence.types import PI, Interaction, Strategy

DigitalEntanglers = Union[CNOT, CZ, CRZ, CRY, CRX, CPHASE]


def _entangler(
    control: int,
    target: int,
    param_str: str,
    entangler: Type[DigitalEntanglers] = CNOT,
) -> AbstractBlock:
    if entangler in [CNOT, CZ]:
        return entangler(control, target)  # type: ignore
    elif entangler in [CRZ, CRY, CRX, CPHASE]:
        param = Parameter(param_str, value=0.0, trainable=True)
        return entangler(control, target, param)  # type: ignore
    else:
        raise ValueError("Provided entangler not accepted for digital ansatz.")


def _entangler_analog(
    param_str: str,
    generator: AbstractBlock | None = None,
) -> AbstractBlock:
    param = Parameter(name=param_str, value=0.0, trainable=True)
    return HamEvo(generator=generator, parameter=param)


def _rotations(
    n_qubits: int,
    layer: int,
    side: str,
    param_str: str,
    values: list[float | torch.Tensor],
    ops: list[type[AbstractBlock]] = [RX, RY],
) -> list[KronBlock]:
    if side == "left":
        idx = lambda x: x  # noqa: E731
    elif side == "right":
        idx = lambda x: len(ops) - x - 1  # noqa: E731
        ops = list(reversed(ops))
    else:
        raise ValueError("Please provide either 'left' or 'right'")

    rot_list = []
    for i, gate in enumerate(ops):
        rot_list.append(
            kron(
                gate(
                    target=n,  # type: ignore [call-arg]
                    parameter=Parameter(  # type: ignore [call-arg]
                        name=param_str + f"_{layer}{n + n_qubits * idx(i)}",
                        value=values[n + n_qubits * idx(i)],
                        trainable=True,
                    ),
                )
                for n in range(n_qubits)
            )
        )

    return rot_list


def identity_initialized_ansatz(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "iia",
    strategy: Strategy = Strategy.DIGITAL,
    rotations: Any = [RX, RY],
    entangler: Any = None,
    periodic: bool = False,
) -> AbstractBlock:
    """
    Identity block for barren plateau mitigation.

    The initial configuration of this block is equal to an identity unitary
    but can be trained in the same fashion as other ansatzes, reaching same level
    of expressivity.

    Args:
        n_qubits: number of qubits in the block
        depth: number of layers of the HEA
        param_prefix (str):
            The base name of the variational parameter. Defaults to "iia".
        strategy: (Strategy)
            Strategy.DIGITAL for fully digital or Strategy.SDAQC for digital-analog.
        rotations (list of AbstractBlocks):
            single-qubit rotations with trainable parameters
        entangler (AbstractBlock):
            For Digital:
                2-qubit entangling operation. Supports CNOT, CZ, CRX, CRY, CRZ, CPHASE.
                Controlled rotations will have variational parameters on the rotation angles.
                Defaults to CNOT.
            For Digital-analog:
                Hamiltonian generator for the analog entangling layer.
                Time parameter is considered variational.
                Defaults to a global NN Hamiltonain.
        periodic (bool): if the qubits should be linked periodically. Valid only for digital.
    """
    initialized_layers = []
    for layer in range(depth):
        alpha = 2 * PI * torch.rand(n_qubits * len(rotations))
        gamma = torch.zeros(n_qubits)
        beta = -alpha

        left_rotations = _rotations(
            n_qubits=n_qubits,
            layer=layer,
            side="left",
            param_str=f"{param_prefix}_α",
            values=alpha,
            ops=rotations,
        )

        if strategy == Strategy.DIGITAL:
            if entangler is None:
                entangler = CNOT

            if entangler not in [CNOT, CZ, CRZ, CRY, CRX, CPHASE]:
                raise ValueError(
                    "Please provide a valid two-qubit entangler operation for digital IIA."
                )

            ent_param_prefix = f"{param_prefix}_θ_ent_"
            if not periodic:
                left_entanglers = [
                    chain(
                        _entangler(
                            control=n,
                            target=n + 1,
                            param_str=ent_param_prefix + f"_{layer}{n}",
                            entangler=entangler,
                        )
                        for n in range(n_qubits - 1)
                    )
                ]
            else:
                left_entanglers = [
                    chain(
                        _entangler(
                            control=n,
                            target=(n + 1) % n_qubits,
                            param_str=ent_param_prefix + f"_{layer}{n}",
                            entangler=entangler,
                        )
                        for n in range(n_qubits)
                    )
                ]

        elif strategy == Strategy.SDAQC:
            if entangler is None:
                entangler = hamiltonian_factory(n_qubits, interaction=Interaction.NN)

            if not block_is_qubit_hamiltonian(entangler):
                raise ValueError(
                    "Please provide a valid Pauli Hamiltonian generator for digital-analog IIA."
                )

            ent_param_prefix = f"{param_prefix}_ent_t"

            left_entanglers = [
                chain(
                    _entangler_analog(
                        param_str=f"{ent_param_prefix}_{layer}",
                        generator=entangler,
                    )
                )
            ]

        else:
            raise NotImplementedError

        centre_rotations = [
            kron(
                RX(
                    target=n,
                    parameter=Parameter(name=f"{param_prefix}_γ" + f"_{layer}{n}", value=gamma[n]),
                )
                for n in range(n_qubits)
            )
        ]

        right_entanglers = reversed(*left_entanglers)

        right_rotations = _rotations(
            n_qubits=n_qubits,
            layer=layer,
            side="right",
            param_str=f"{param_prefix}_β",
            values=beta,
            ops=rotations,
        )

        krons = [
            *left_rotations,
            *left_entanglers,
            *centre_rotations,
            *right_entanglers,
            *right_rotations,
        ]

        initialized_layers.append(tag(chain(*krons), tag=f"BPMA-{layer}"))

    return chain(*initialized_layers)



from __future__ import annotations

from typing import Callable, List, Type, Union

import numpy as np
from torch import Tensor, double, ones, rand

from qadence.blocks import AbstractBlock, add, block_is_qubit_hamiltonian
from qadence.logger import get_logger
from qadence.operations import N, X, Y, Z
from qadence.register import Register
from qadence.types import Interaction, TArray

logger = get_logger(__name__)


def interaction_zz(i: int, j: int) -> AbstractBlock:
    """Ising ZZ interaction."""
    return Z(i) @ Z(j)


def interaction_nn(i: int, j: int) -> AbstractBlock:
    """Ising NN interaction."""
    return N(i) @ N(j)


def interaction_xy(i: int, j: int) -> AbstractBlock:
    """XY interaction."""
    return X(i) @ X(j) + Y(i) @ Y(j)


def interaction_xyz(i: int, j: int) -> AbstractBlock:
    """Heisenberg XYZ interaction."""
    return X(i) @ X(j) + Y(i) @ Y(j) + Z(i) @ Z(j)


INTERACTION_DICT = {
    Interaction.ZZ: interaction_zz,
    Interaction.NN: interaction_nn,
    Interaction.XY: interaction_xy,
    Interaction.XYZ: interaction_xyz,
}


ARRAYS = (list, np.ndarray, Tensor)

DETUNINGS = (N, X, Y, Z)

TDetuning = Union[Type[N], Type[X], Type[Y], Type[Z]]


def hamiltonian_factory(
    register: Register | int,
    interaction: Interaction | Callable | None = None,
    detuning: TDetuning | None = None,
    interaction_strength: TArray | str | None = None,
    detuning_strength: TArray | str | None = None,
    random_strength: bool = False,
    use_all_node_pairs: bool = False,
) -> AbstractBlock:
    """
    General Hamiltonian creation function.

    Can be used to create Hamiltonians with 2-qubit
    interactions and single-qubit detunings, both with arbitrary strength or parameterized.

    Arguments:
        register: register of qubits with a specific graph topology, or number of qubits.
            When passing a number of qubits a register with all-to-all connectivity
            is created.
        interaction: Interaction.ZZ, Interaction.NN, Interaction.XY, or Interacton.XYZ.
        detuning: single-qubit operator N, X, Y, or Z.
        interaction_strength: list of values to be used as the interaction strength for each
            pair of qubits. Should be ordered following the order of `Register(n_qubits).edges`.
            Alternatively, some string "x" can be passed, which will create a parameterized
            interactions for each pair of qubits, each labelled as `"x_ij"`.
        detuning_strength: list of values to be used as the detuning strength for each qubit.
            Alternatively, some string "x" can be passed, which will create a parameterized
            detuning for each qubit, each labelled as `"x_i"`.
        random_strength: set random interaction and detuning strengths between -1 and 1.
        use_all_node_pairs: computes an interaction term for every pair of nodes in the graph,
            independent of the edge topology in the register. Useful for defining Hamiltonians
            where the interaction strength decays with the distance.

    Examples:
        ```python exec="on" source="material-block" result="json"
        from qadence import hamiltonian_factory, Interaction, Register, Z

        n_qubits = 3

        # Constant total magnetization observable:
        observable = hamiltonian_factory(n_qubits, detuning = Z)

        # Parameterized total magnetization observable:
        observable = hamiltonian_factory(n_qubits, detuning = Z, detuning_strength = "z")

        # Random all-to-all XY Hamiltonian generator:
        generator = hamiltonian_factory(
            n_qubits,
            interaction = Interaction.XY,
            random_strength = True,
            )

        # Parameterized NN Hamiltonian generator with a square grid interaction topology:
        register = Register.square(qubits_side = n_qubits)
        generator = hamiltonian_factory(
            register,
            interaction = Interaction.NN,
            interaction_strength = "theta"
            )
        ```
    """

    if interaction is None and detuning is None:
        raise ValueError("Please provide an interaction and/or detuning for the Hamiltonian.")

    # If number of qubits is given, creates all-to-all register
    register = Register(register) if isinstance(register, int) else register

    # Get interaction function
    if interaction is not None:
        if callable(interaction):
            int_fn = interaction
            try:
                if not block_is_qubit_hamiltonian(interaction(0, 1)):
                    raise ValueError("Custom interactions must be composed of Pauli operators.")
            except TypeError:
                raise TypeError(
                    "Please use a custom interaction function signed with two integer parameters."
                )
        else:
            int_fn = INTERACTION_DICT.get(interaction, None)  # type: ignore [arg-type]
            if int_fn is None:
                raise KeyError(f"Interaction {interaction} not supported.")

    # Check single-qubit detuning
    if (detuning is not None) and (detuning not in DETUNINGS):
        raise TypeError(f"Detuning of type {type(detuning)} not supported.")

    # Pre-process detuning and interaction strengths and update register
    detuning_strength_array = _preprocess_strengths(
        register, detuning_strength, "nodes", random_strength
    )

    edge_str = "all_node_pairs" if use_all_node_pairs else "edges"
    interaction_strength_array = _preprocess_strengths(
        register, interaction_strength, edge_str, random_strength
    )

    # Create single-qubit detunings:
    single_qubit_terms: List[AbstractBlock] = []
    if detuning is not None:
        for strength, node in zip(detuning_strength_array, register.nodes):
            single_qubit_terms.append(strength * detuning(node))

    # Create two-qubit interactions:
    two_qubit_terms: List[AbstractBlock] = []
    edge_data = register.all_node_pairs if use_all_node_pairs else register.edges
    if interaction is not None and int_fn is not None:
        for strength, edge in zip(interaction_strength_array, edge_data):
            two_qubit_terms.append(strength * int_fn(*edge))

    return add(*single_qubit_terms, *two_qubit_terms)


def _preprocess_strengths(
    register: Register,
    strength: TArray | str | None,
    nodes_or_edges: str,
    random_strength: bool,
) -> Tensor | list:
    data = getattr(register, nodes_or_edges)

    # Useful for error messages:
    strength_target = "detuning" if nodes_or_edges == "nodes" else "interaction"

    # Next we process the strength given in the input arguments
    if strength is None:
        if random_strength:
            strength = 2 * rand(len(data), dtype=double) - 1
        else:
            # None defaults to constant = 1.0
            strength = ones(len(data), dtype=double)
    elif isinstance(strength, ARRAYS):
        # If array is given, checks it has the correct length
        if len(strength) != len(data):
            message = "Array of " + strength_target + " strengths has incorrect size."
            raise ValueError(message)
    elif isinstance(strength, str):
        prefix = strength
        if nodes_or_edges == "nodes":
            strength = [prefix + f"_{node}" for node in data]
        if nodes_or_edges in ["edges", "all_node_pairs"]:
            strength = [prefix + f"_{edge[0]}{edge[1]}" for edge in data]
    else:
        # If not of the accepted types ARRAYS or str, we error out
        raise TypeError(
            "Incorrect " + strength_target + f" strength type {type(strength)}. "
            "Please provide an array of strength values, or a string for "
            "parameterized " + strength_target + "s."
        )

    return strength


def total_magnetization(n_qubits: int, z_terms: np.ndarray | list | None = None) -> AbstractBlock:
    return hamiltonian_factory(n_qubits, detuning=Z, detuning_strength=z_terms)


def single_z(qubit: int = 0, z_coefficient: float = 1.0) -> AbstractBlock:
    return Z(qubit) * z_coefficient


def zz_hamiltonian(
    n_qubits: int,
    z_terms: np.ndarray | None = None,
    zz_terms: np.ndarray | None = None,
) -> AbstractBlock:
    if zz_terms is not None:
        register = Register(n_qubits)
        interaction_strength = [zz_terms[edge[0], edge[1]] for edge in register.edges]
    else:
        interaction_strength = None
    return hamiltonian_factory(n_qubits, Interaction.ZZ, Z, interaction_strength, z_terms)


def ising_hamiltonian(
    n_qubits: int,
    x_terms: np.ndarray | None = None,
    z_terms: np.ndarray | None = None,
    zz_terms: np.ndarray | None = None,
) -> AbstractBlock:
    zz_ham = zz_hamiltonian(n_qubits, z_terms=z_terms, zz_terms=zz_terms)
    x_ham = hamiltonian_factory(n_qubits, detuning=X, detuning_strength=x_terms)
    return zz_ham + x_ham



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

from qadence.backend import BackendConfiguration
from qadence.logger import get_logger
from qadence.transpile import (
    blockfn_to_circfn,
    flatten,
    scale_primitive_blocks_only,
)

logger = get_logger(__name__)


def default_passes(config: Configuration) -> list[Callable]:
    return [
        flatten,
        blockfn_to_circfn(scale_primitive_blocks_only),
    ]


@dataclass
class Configuration(BackendConfiguration):
    pass



from __future__ import annotations

from dataclasses import dataclass
from itertools import chain as flatten
from pathlib import Path
from typing import Iterable

from sympy import Basic

from qadence.blocks import AbstractBlock, AnalogBlock, CompositeBlock, chain
from qadence.blocks.utils import parameters, primitive_blocks, unique_parameters
from qadence.parameters import Parameter
from qadence.register import Register

# Modules to be automatically added to the qadence namespace
__all__ = ["QuantumCircuit"]


@dataclass(eq=False)  # Avoid unhashability errors due to mutable attributes.
class QuantumCircuit:
    """Am abstract QuantumCircuit instance.

    It needs to be passed to a quantum backend for execution.
    """

    block: AbstractBlock
    register: Register

    def __init__(self, support: int | Register, *blocks: AbstractBlock):
        """
        Arguments:

            support: `Register` or number of qubits. If an integer is provided, a register is
                constructed with `Register.all_to_all(x)`
            *blocks: (Possibly multiple) blocks to construct the circuit from.
        """
        self.block = chain(*blocks) if len(blocks) != 1 else blocks[0]
        self.register = Register(support) if isinstance(support, int) else support

        global_block = isinstance(self.block, AnalogBlock) and self.block.qubit_support.is_global
        if not global_block and len(self.block) and self.block.n_qubits > self.register.n_qubits:
            raise ValueError(
                f"Register with {self.register.n_qubits} qubits is too small for the "
                f"given block with {self.block.n_qubits} qubits"
            )

    @property
    def n_qubits(self) -> int:
        return self.register.n_qubits

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QuantumCircuit):
            raise TypeError(f"Cannot compare {type(self)} to {type(other)}.")
        if self.block != other.block:  # type: ignore[call-overload]
            return False
        if self.register != other.register:
            return False
        return True

    def __hash__(self) -> int:
        return hash(self._to_json())

    def __iter__(self) -> Iterable:
        if isinstance(self.block, CompositeBlock):
            yield from self.block
        else:
            yield self.block

    def __contains__(self, other: object) -> bool:
        if isinstance(other, AbstractBlock):
            if isinstance(self.block, CompositeBlock):
                return other in self.block
            else:
                return other == self.block
        elif isinstance(other, Parameter):
            return other in self.unique_parameters
        else:
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")

    @property
    def unique_parameters(self) -> list[Parameter]:
        """Return the unique parameters in the circuit.

        These parameters are the actual user-facing parameters which
        can be assigned by the user. Multiple gates can contain the
        same unique parameter

        Returns:
            list[Parameter]: List of unique parameters in the circuit
        """
        return unique_parameters(self.block)

    @property
    def num_unique_parameters(self) -> int:
        return len(self.unique_parameters) if self.unique_parameters else 0

    @property
    def num_parameters(self) -> int:
        return len(self.parameters())

    def parameters(self) -> list[Parameter | Basic] | list[tuple[Parameter | Basic, ...]]:
        """Extract all parameters for primitive blocks in the circuit.

        Notice that this function returns all the unique Parameters used
        in the quantum circuit. These can correspond to constants too.

        Returns:
            List[tuple[Parameter]]: A list of tuples containing the Parameter
            instance of each of the primitive blocks in the circuit or, if the `flatten`
            flag is set to True, a flattened list of all circuit parameters
        """
        return parameters(self.block)

    def dagger(self) -> QuantumCircuit:
        """Reverse the QuantumCircuit by calling dagger on the block."""
        return QuantumCircuit(self.n_qubits, self.block.dagger())

    def get_blocks_by_tag(self, tag: str) -> list[AbstractBlock]:
        """Extract one or more blocks using the human-readable tag.

        This function recursively explores all composite blocks to find
        all the occurrences of a certain tag in the blocks.

        Args:
            tag (str): the tag to look for

        Returns:
            list[AbstractBlock]: The block(s) corresponding to the given tag
        """

        def _get_block(block: AbstractBlock) -> list[AbstractBlock]:
            blocks = []
            if block.tag == tag:
                blocks += [block]
            if isinstance(block, CompositeBlock):
                blocks += flatten(*[_get_block(b) for b in block.blocks])
            return blocks

        return _get_block(self.block)

    def is_empty(self) -> bool:
        return len(primitive_blocks(self.block)) == 0

    def serialize(self) -> str:
        raise NotImplementedError

    @staticmethod
    def deserialize(json: str) -> QuantumCircuit:
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.block.__repr__()

    def _to_dict(self) -> dict:
        return {
            "block": self.block._to_dict(),
            "register": self.register._to_dict(),
        }

    def _to_json(self, path: Path | str | None = None) -> str:
        import json

        qc_dumped = json.dumps(self._to_dict())
        if path is not None:
            path = Path(path)
            try:
                with open(path, "w") as file:
                    file.write(qc_dumped)
            except Exception as e:
                print(f"Unable to write QuantumCircuit to disk due to {e}")

        return qc_dumped

    @classmethod
    def _from_dict(cls, d: dict) -> QuantumCircuit:
        from qadence import blocks as qadenceblocks
        from qadence import operations

        RootBlock = (
            getattr(operations, d["block"]["type"])
            if hasattr(operations, d["block"]["type"])
            else getattr(qadenceblocks, d["block"]["type"])
        )

        return QuantumCircuit(
            Register._from_dict(d["register"]),
            RootBlock._from_dict(d["block"]),
        )

    @classmethod
    def _from_json(cls, path: str | Path) -> QuantumCircuit:
        import json

        loaded_dict: dict = {}
        if isinstance(path, str):
            path = Path(path)
        try:
            with open(path, "r") as file:
                loaded_dict = json.load(file)

        except Exception as e:
            print(f"Unable to load QuantumCircuit due to {e}")

        return QuantumCircuit._from_dict(loaded_dict)



# flake8: noqa
import warnings
from typing import Any

from qadence.register import Register
from qadence.blocks.analog import AnalogBlock, Interaction

from .abstract import AbstractBlock
from .primitive import (
    ParametricBlock,
    PrimitiveBlock,
    TimeEvolutionBlock,
    ScaleBlock,
    ParametricControlBlock,
    ControlBlock,
)
from .composite import AddBlock, ChainBlock, CompositeBlock, KronBlock, PutBlock
from .matrix import MatrixBlock
from .manipulate import from_openfermion, to_openfermion
from .utils import (
    add,
    chain,
    kron,
    tag,
    put,
    block_is_commuting_hamiltonian,
    block_is_qubit_hamiltonian,
    parameters,
    primitive_blocks,
    get_pauli_blocks,
    has_duplicate_vparams,
)
from .block_to_tensor import block_to_tensor
from .embedding import embedding

# Modules to be automatically added to the qadence namespace
__all__ = ["add", "chain", "kron", "tag", "block_to_tensor"]



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from typing import Any

import pyqtorch as pyq
import torch
from torch import Tensor

from qadence.backend import Backend as BackendInterface
from qadence.backend import ConvertedCircuit, ConvertedObservable
from qadence.backends.utils import (
    pyqify,
    to_list_of_dicts,
    unpyqify,
    validate_state,
)
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.logger import get_logger
from qadence.measurements import Measurements
from qadence.mitigations.protocols import Mitigations, apply_mitigation
from qadence.noise import Noise
from qadence.noise.protocols import apply_noise
from qadence.transpile import (
    chain_single_qubit_ops,
    flatten,
    invert_endianness,
    scale_primitive_blocks_only,
    transpile,
)
from qadence.types import BackendName, Endianness, Engine
from qadence.utils import infer_batchsize, int_to_basis

from .config import Configuration, default_passes
from .convert_ops import convert_block, convert_observable

logger = get_logger(__name__)


@dataclass(frozen=True, eq=True)
class Backend(BackendInterface):
    """PyQTorch backend."""

    name: BackendName = BackendName.PYQTORCH
    supports_ad: bool = True
    support_bp: bool = True
    supports_adjoint: bool = True
    is_remote: bool = False
    with_measurements: bool = True
    with_noise: bool = False
    native_endianness: Endianness = Endianness.BIG
    config: Configuration = field(default_factory=Configuration)
    engine: Engine = Engine.TORCH

    def circuit(self, circuit: QuantumCircuit) -> ConvertedCircuit:
        passes = self.config.transpilation_passes
        if passes is None:
            passes = default_passes(self.config)

        original_circ = circuit
        if len(passes) > 0:
            circuit = transpile(*passes)(circuit)

        ops = convert_block(circuit.block, n_qubits=circuit.n_qubits, config=self.config)
        native = pyq.QuantumCircuit(circuit.n_qubits, ops)
        return ConvertedCircuit(native=native, abstract=circuit, original=original_circ)

    def observable(self, observable: AbstractBlock, n_qubits: int) -> ConvertedObservable:
        # make sure only leaves, i.e. primitive blocks are scaled
        transpilations = [
            lambda block: chain_single_qubit_ops(block)
            if self.config.use_single_qubit_composition
            else flatten(block),
            scale_primitive_blocks_only,
        ]
        block = transpile(*transpilations)(observable)  # type: ignore[call-overload]

        (native,) = convert_observable(block, n_qubits=n_qubits, config=self.config)
        return ConvertedObservable(native=native, abstract=block, original=observable)

    def _run(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
        pyqify_state: bool = True,
        unpyqify_state: bool = True,
    ) -> Tensor:
        n_qubits = circuit.abstract.n_qubits
        if state is None:
            # If no state is passed, we infer the batch_size through the length
            # of the individual parameter value tensors.
            state = circuit.native.init_state(batch_size=infer_batchsize(param_values))
        else:
            validate_state(state, n_qubits)
            # pyqtorch expects input shape [2] * n_qubits + [batch_size]
            state = pyqify(state, n_qubits) if pyqify_state else state
        state = circuit.native.run(state, param_values)
        state = unpyqify(state) if unpyqify_state else state
        state = invert_endianness(state) if endianness != self.native_endianness else state
        return state

    def run_dm(
        self,
        circuit: ConvertedCircuit,
        noise: Noise,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        raise NotImplementedError

    def _batched_expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        state = self.run(
            circuit,
            param_values=param_values,
            state=state,
            endianness=endianness,
            pyqify_state=True,
            # we are calling  the native observable directly, so we want to use pyq shapes
            unpyqify_state=False,
        )
        observable = observable if isinstance(observable, list) else [observable]
        _expectation = torch.hstack(
            [obs.native(state, param_values).reshape(-1, 1) for obs in observable]
        )
        return _expectation

    def _looped_expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        if state is None:
            from qadence.states import zero_state

            state = zero_state(circuit.abstract.n_qubits, batch_size=1)
        if state.size(0) != 1:
            raise ValueError(
                "Looping expectation does not make sense with batched initial state. "
                "Define your initial state with `batch_size=1`"
            )

        list_expvals = []
        observables = observable if isinstance(observable, list) else [observable]
        for vals in to_list_of_dicts(param_values):
            wf = self.run(circuit, vals, state, endianness, pyqify_state=True, unpyqify_state=False)
            exs = torch.cat([obs.native(wf, vals) for obs in observables], 0)
            list_expvals.append(exs)

        batch_expvals = torch.vstack(list_expvals)
        return batch_expvals if len(batch_expvals.shape) > 0 else batch_expvals.reshape(1)

    def expectation(
        self,
        circuit: ConvertedCircuit,
        observable: list[ConvertedObservable] | ConvertedObservable,
        param_values: dict[str, Tensor] = {},
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        # Noise is ignored if measurement protocol is not provided.
        if noise is not None and measurement is None:
            logger.warning(
                f"Errors of type {noise} are not implemented for exact expectation yet. "
                "This is ignored for now."
            )
        fn = self._looped_expectation if self.config.loop_expectation else self._batched_expectation
        return fn(
            circuit=circuit,
            observable=observable,
            param_values=param_values,
            state=state,
            measurement=measurement,
            noise=noise,
            endianness=endianness,
        )

    def sample(
        self,
        circuit: ConvertedCircuit,
        param_values: dict[str, Tensor] = {},
        n_shots: int = 1,
        state: Tensor | None = None,
        noise: Noise | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        if n_shots < 1:
            raise ValueError("You can only call sample with n_shots>0.")

        def _sample(_probs: Tensor, n_shots: int, endianness: Endianness, n_qubits: int) -> Counter:
            return Counter(
                {
                    int_to_basis(k=k, n_qubits=n_qubits, endianness=endianness): count.item()
                    for k, count in enumerate(
                        torch.bincount(
                            torch.multinomial(input=_probs, num_samples=n_shots, replacement=True)
                        )
                    )
                    if count > 0
                }
            )

        wf = self.run(circuit=circuit, param_values=param_values, state=state)
        probs = torch.abs(torch.pow(wf, 2))
        samples = list(
            map(
                lambda _probs: _sample(
                    _probs=_probs,
                    n_shots=n_shots,
                    endianness=endianness,
                    n_qubits=circuit.abstract.n_qubits,
                ),
                probs,
            )
        )
        if noise is not None:
            samples = apply_noise(noise=noise, samples=samples)
        if mitigation is not None:
            assert noise
            samples = apply_mitigation(noise=noise, mitigation=mitigation, samples=samples)
        return samples

    def assign_parameters(self, circuit: ConvertedCircuit, param_values: dict[str, Tensor]) -> Any:
        raise NotImplementedError

    @staticmethod
    def _overlap(bras: Tensor, kets: Tensor) -> Tensor:
        from qadence.overlap import overlap_exact

        return overlap_exact(bras, kets)

    @staticmethod
    def default_configuration() -> Configuration:
        return Configuration()



from __future__ import annotations

from sympy import cos, sin
from torch import float64, tensor

from qadence.analog.constants import C6_DICT
from qadence.blocks import add
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.analog import ConstantAnalogRotation
from qadence.constructors import hamiltonian_factory
from qadence.operations import I, N, X, Y, Z
from qadence.register import Register
from qadence.types import Interaction


def rydberg_interaction_hamiltonian(
    register: Register,
) -> AbstractBlock:
    """
    Computes the Rydberg Ising or XY interaction Hamiltonian for a register of qubits.

    H_int = ∑_(j<i) (C_6 / R**6) * kron(N_i, N_j)

    H_int = ∑_(j<i) (C_3 / R**3) * (kron(X_i, X_j) + kron(Y_i, Y_j))

    Args:
        register: the register of qubits.
    """

    distances = tensor(list(register.distances.values()), dtype=float64)
    device_specs = register.device_specs

    if device_specs.interaction == Interaction.NN:
        c6 = C6_DICT[device_specs.rydberg_level]
        strength_list = c6 / (distances**6)
    elif device_specs.interaction == Interaction.XY:
        c3 = device_specs.coeff_xy
        strength_list = c3 / (distances**3)

    return hamiltonian_factory(
        register,
        interaction=device_specs.interaction,
        interaction_strength=strength_list,
        use_all_node_pairs=True,
    )


def rydberg_drive_hamiltonian(block: ConstantAnalogRotation, register: Register) -> AbstractBlock:
    """
    Computes the Rydberg drive Hamiltonian for some local or global rotation.

    H_d = ∑_i (Ω/2 cos(φ) * X_i - sin(φ) * Y_i) - δ * N_i

    Args:
        block: the ConstantAnalogRotation block containing the parameters.
        register: the register of qubits.
    """

    if block.qubit_support.is_global:
        qubit_support = tuple(register.nodes)
    else:
        qubit_support = block.qubit_support

    omega = block.parameters.omega
    delta = block.parameters.delta
    phase = block.parameters.phase

    x_terms = (omega / 2) * add(cos(phase) * X(i) for i in qubit_support)
    y_terms = (omega / 2) * add(sin(phase) * Y(i) for i in qubit_support)
    n_terms = delta * add(N(i) for i in qubit_support)
    h_drive: AbstractBlock = x_terms - y_terms - n_terms
    return h_drive


def rydberg_pattern_hamiltonian(register: Register) -> AbstractBlock | None:
    support = tuple(range(register.n_qubits))
    pattern = register.device_specs.pattern
    if pattern is not None:
        amp = pattern.amp
        det = pattern.det
        weights_amp = pattern.weights_amp
        weights_det = pattern.weights_det
        local_constr_amp = pattern.local_constr_amp
        local_constr_det = pattern.local_constr_det
        global_constr_amp = pattern.global_constr_amp
        global_constr_det = pattern.global_constr_det

        p_amp_terms: AbstractBlock = (
            (1 / 2)  # type: ignore [operator]
            * amp
            * global_constr_amp
            * add(X(i) * weights_amp[i] * local_constr_amp[i] for i in support)  # type: ignore [operator]
        )
        p_det_terms: AbstractBlock = (
            -det  # type: ignore [operator]
            * global_constr_det
            * add(0.5 * (I(i) - Z(i)) * weights_det[i] * local_constr_det[i] for i in support)  # type: ignore [operator]
        )
        return p_amp_terms + p_det_terms
    else:
        return None



from __future__ import annotations

from collections import Counter
from functools import reduce

import numpy as np
import torch
from torch import Tensor

from qadence.backend import Backend
from qadence.backends.pyqtorch import Backend as PyQBackend
from qadence.blocks import AbstractBlock, PrimitiveBlock, chain
from qadence.blocks.utils import unroll_block_with_scaling
from qadence.circuit import QuantumCircuit
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.noise import Noise
from qadence.operations import H, SDagger, X, Y, Z
from qadence.parameters import evaluate
from qadence.utils import Endianness


def get_qubit_indices_for_op(pauli_term: tuple, op: PrimitiveBlock | None = None) -> list[int]:
    """Get qubit indices for the given op in the Pauli term if any."""
    indices = []
    blocks = getattr(pauli_term[0], "blocks", None)
    if blocks is not None:
        for block in blocks:
            if op is None:
                indices.append(block.qubit_support[0])
            if isinstance(block, type(op)):
                indices.append(block.qubit_support[0])
    else:
        block = pauli_term[0]
        if op is None:
            indices.append(block.qubit_support[0])
        if isinstance(block, type(op)):
            indices.append(block.qubit_support[0])
    return indices


def rotate(circuit: QuantumCircuit, pauli_term: tuple) -> QuantumCircuit:
    """Rotate circuit to measurement basis and return the qubit support."""
    rotations = []

    # Mypy expects concrete types. Although there definitely should be
    # a better way to pass the operation type.
    for op, gate in [(X(0), Z), (Y(0), SDagger)]:
        qubit_indices = get_qubit_indices_for_op(pauli_term, op=op)
        for index in qubit_indices:
            rotations.append(gate(index) * H(index))
    rotated_block = chain(circuit.block, *rotations)
    return QuantumCircuit(circuit.register, rotated_block)


def get_counts(samples: list, support: list) -> list:
    """Marginalise the probablity mass function to the support."""
    counts = []
    for sample in samples:
        sample_counts = []
        for k, v in sample.items():
            sample_counts.append(Counter({"".join([k[i] for i in support]): sample[k]}))
        reduced_counts = reduce(lambda x, y: x + y, sample_counts)
        counts.append(reduced_counts)
    return counts


def empirical_average(samples: list, support: list) -> Tensor:
    """Compute the empirical average."""
    counters = get_counts(samples, support)
    expectations = []
    n_shots = np.sum(list(counters[0].values()))
    parity = -1
    for counter in counters:
        counter_exps = []
        for bitstring, count in counter.items():
            counter_exps.append(count * parity ** (np.sum([int(bit) for bit in bitstring])))
        expectations.append(np.sum(counter_exps) / n_shots)
    return torch.tensor(expectations)


def iterate_pauli_decomposition(
    circuit: QuantumCircuit,
    param_values: dict,
    pauli_decomposition: list,
    n_shots: int,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Estimate total expectation value by averaging all Pauli terms."""

    estimated_values = []

    for pauli_term in pauli_decomposition:
        if pauli_term[0].is_identity:
            estimated_values.append(evaluate(pauli_term[1], as_torch=True))
        else:
            # Get the full qubit support for the Pauli term.
            # Note: duplicates must be kept here to allow for
            # observables chaining multiple operations on the same qubit
            # such as `b = chain(Z(0), Z(0))`
            support = get_qubit_indices_for_op(pauli_term)
            # Rotate the circuit according to the given observable term.
            rotated_circuit = rotate(circuit=circuit, pauli_term=pauli_term)
            # Use the low-level backend API to avoid embedding of parameters
            # already performed at the higher QuantumModel level.
            # Therefore, parameters passed here have already been embedded.
            conv_circ = backend.circuit(rotated_circuit)
            samples = backend.sample(
                circuit=conv_circ,
                param_values=param_values,
                n_shots=n_shots,
                state=state,
                noise=noise,
                endianness=endianness,
            )
            estim_values = empirical_average(samples=samples, support=support)
            # TODO: support for parametric observables to be tested
            estimated_values.append(estim_values * evaluate(pauli_term[1]))
    res = torch.sum(torch.stack(estimated_values), axis=0)
    # Allow for automatic differentiation.
    res.requires_grad = True
    return res


def compute_expectation(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    options: dict,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Basic tomography protocol with rotations.

    Given a circuit and a list of observables, apply basic tomography protocol to estimate
    the expectation values.

    Args:
        circuit (QuantumCircuit): a circuit to prepare the state.
        observables (list[AbstractBlock]): a list of observables
            to estimate the expectation values from.
        param_values (dict): a dict of values to substitute the
            symbolic parameters for.
        options (dict): a dict of options for the measurement protocol.
            Here, shadow_size (int), accuracy (float) and confidence (float) are supported.
        state (Tensor | None): an initial input state.
        backend_name (BackendName): a backend name to retrieve computations from.
        noise: A noise model to use.
        endianness: Endianness of the observable estimate.
    """
    if not isinstance(observables, list):
        raise TypeError(
            "Observables must be of type <class 'List[AbstractBlock]'>. Got {}.".format(
                type(observables)
            )
        )
    n_shots = options.get("n_shots")
    if n_shots is None:
        raise KeyError("Tomography protocol requires a 'n_shots' kwarg of type 'int'.")
    estimated_values = []
    for observable in observables:
        pauli_decomposition = unroll_block_with_scaling(observable)
        estimated_values.append(
            iterate_pauli_decomposition(
                circuit=circuit,
                param_values=param_values,
                pauli_decomposition=pauli_decomposition,
                n_shots=n_shots,
                state=state,
                backend=backend,
                noise=noise,
                endianness=endianness,
            )
        )
    return torch.transpose(torch.vstack(estimated_values), 1, 0)



from __future__ import annotations

from typing import Callable

import numpy as np
from sympy import Basic

from qadence.blocks import AnalogBlock, KronBlock, kron
from qadence.constructors.feature_maps import fm_parameter_func, fm_parameter_scaling
from qadence.logger import get_logger
from qadence.operations import AnalogRot, AnalogRX, AnalogRY, AnalogRZ
from qadence.parameters import FeatureParameter, Parameter, VariationalParameter
from qadence.types import PI, BasisSet, ReuploadScaling, TParameter

logger = get_logger(__file__)

AnalogRotationTypes = [AnalogRX, AnalogRY, AnalogRZ]


def rydberg_feature_map(
    n_qubits: int,
    param: str = "phi",
    max_abs_detuning: float = 2 * PI * 10,
    weights: list[float] | None = None,
) -> KronBlock:
    """Feature map using semi-local addressing patterns.

    If not weights are specified, variational parameters are created
    for the pattern

    Args:
        n_qubits (int): number of qubits
        param: the name of the feature parameter
        max_abs_detuning: maximum value of absolute detuning for each qubit. Defaulted at 10 MHz.
        weights: a list of wegiths to assign to each qubit parameter in the feature map

    Returns:
        The block representing the feature map
    """

    tower_coeffs: list[float | Parameter]
    tower_coeffs = (
        [VariationalParameter(f"w_{param}_{i}") for i in range(n_qubits)]
        if weights is None
        else weights
    )
    tower_detuning = max_abs_detuning / (sum(tower_coeffs[i] for i in range(n_qubits)))

    param = FeatureParameter(param)
    duration = 1000 * param / tower_detuning
    return kron(
        AnalogRot(
            duration=duration,
            delta=-tower_detuning * tower_coeffs[i],
            phase=0.0,
            qubit_support=(i,),
        )
        for i in range(n_qubits)
    )


def rydberg_tower_feature_map(
    n_qubits: int, param: str = "phi", max_abs_detuning: float = 2 * PI * 10
) -> KronBlock:
    weights = list(np.arange(1, n_qubits + 1))
    return rydberg_feature_map(
        n_qubits, param=param, max_abs_detuning=max_abs_detuning, weights=weights
    )


def analog_feature_map(
    param: str = "phi",
    op: Callable[[Parameter | Basic], AnalogBlock] = AnalogRX,
    fm_type: BasisSet | Callable | str = BasisSet.FOURIER,
    reupload_scaling: ReuploadScaling | Callable | str = ReuploadScaling.CONSTANT,
    feature_range: tuple[float, float] | None = None,
    target_range: tuple[float, float] | None = None,
    multiplier: Parameter | TParameter | None = None,
) -> AnalogBlock:
    """Generate a fully analog feature map.

    Args:
        param: Parameter of the feature map; you can pass a string or Parameter;
            it will be set as non-trainable (FeatureParameter) regardless.
        op: type of operation. Choose among AnalogRX, AnalogRY, AnalogRZ or a custom
            callable function returning an AnalogBlock instance
        fm_type: Basis set for data encoding; choose from `BasisSet.FOURIER` for Fourier
            encoding, or `BasisSet.CHEBYSHEV` for Chebyshev polynomials of the first kind.
        reupload_scaling: how the feature map scales the data that is re-uploaded. Given that
            this feature map uses analog rotations, the reuploading works by simply
            adding additional operations with different scaling factors in the parameter.
            Choose from `ReuploadScaling` enumeration, currently only CONSTANT works,
            or provide your own function with the first argument being the given
            operation `op` and the second argument the feature parameter
        feature_range: range of data that the input data is assumed to come from.
        target_range: range of data the data encoder assumes as the natural range. For example,
            in Chebyshev polynomials it is (-1, 1), while for Fourier it may be chosen as (0, 2*PI).
        multiplier: overall multiplier; this is useful for reuploading the feature map serially with
            different scalings; can be a number or parameter/expression.
    """

    scaled_fparam = fm_parameter_scaling(
        fm_type, param, feature_range=feature_range, target_range=target_range
    )

    transform_func = fm_parameter_func(fm_type)

    transformed_feature = transform_func(scaled_fparam)

    multiplier = 1.0 if multiplier is None else Parameter(multiplier)

    if callable(reupload_scaling):
        return reupload_scaling(op, multiplier * transformed_feature)  # type: ignore[no-any-return]
    elif reupload_scaling == ReuploadScaling.CONSTANT:
        return op(multiplier * transformed_feature)
    # TODO: implement tower scaling by reuploading multiple times
    # using different analog rotations
    else:
        raise NotImplementedError(f"Reupload scaling {str(reupload_scaling)} not implemented!")



from __future__ import annotations

import json
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import ClassVar, Iterable, Tuple, TypeVar, Union, get_args

import sympy
import torch
from rich.console import Console, RenderableType
from rich.tree import Tree

from qadence.parameters import Parameter
from qadence.types import TNumber


@dataclass(eq=False)  # Avoid unhashability errors due to mutable attributes.
class AbstractBlock(ABC):
    """Base class for both primitive and composite blocks.

    Attributes:
        name (str): A human-readable name attached to the block type. Notice, this is
            the same for all the class instances so it cannot be used for identifying
            different blocks
        qubit_support (tuple[int, ...]): The qubit support of the block expressed as
            a tuple of integers
        tag (str | None): A tag identifying a particular instance of the block which can
            be used for identification and pretty printing
        eigenvalues (list[float] | None): The eigenvalues of the matrix representing the block.
            This is used mainly for primitive blocks and it's needed for generalized parameter
            shift rule computations. Currently unused.
    """

    name: ClassVar[str] = "AbstractBlock"
    tag: str | None = None

    @abstractproperty
    def qubit_support(self) -> Tuple[int, ...]:
        """The indices of the qubit(s) the block is acting on.

        Qadence uses the ordering [0..,N-1] for qubits.
        """
        pass

    @abstractproperty
    def n_qubits(self) -> int:
        """The number of qubits in the whole system.

        A block acting on qubit N would has at least n_qubits >= N + 1.
        """
        pass

    @abstractproperty
    def n_supports(self) -> int:
        """The number of qubits the block is acting on."""
        pass

    @cached_property
    @abstractproperty
    def eigenvalues_generator(self) -> torch.Tensor:
        pass

    @cached_property
    def eigenvalues(
        self, max_num_evals: int | None = None, max_num_gaps: int | None = None
    ) -> torch.Tensor:
        from qadence.utils import eigenvalues

        from .block_to_tensor import block_to_tensor

        return eigenvalues(block_to_tensor(self), max_num_evals, max_num_gaps)

    # make sure that __rmul__ works as expected with np.number
    __array_priority__: int = 1000

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass

    def __mul__(self, other: Union[AbstractBlock, TNumber, Parameter]) -> AbstractBlock:
        from qadence.blocks.primitive import ScaleBlock
        from qadence.blocks.utils import chain

        # TODO: Improve type checking here
        if isinstance(other, AbstractBlock):
            return chain(self, other)
        else:
            if isinstance(self, ScaleBlock):
                scale = self.parameters.parameter * other
                return ScaleBlock(self.block, scale)
            else:
                scale = Parameter(other)
                return ScaleBlock(self, scale)

    def __rmul__(self, other: AbstractBlock | TNumber | Parameter) -> AbstractBlock:
        return self.__mul__(other)

    def __imul__(self, other: Union[AbstractBlock, TNumber, Parameter]) -> AbstractBlock:
        from qadence.blocks.composite import ChainBlock
        from qadence.blocks.primitive import ScaleBlock
        from qadence.blocks.utils import chain

        if not isinstance(other, AbstractBlock):
            raise TypeError("In-place multiplication is available only for AbstractBlock instances")

        # TODO: Improve type checking here
        if isinstance(other, AbstractBlock):
            return chain(
                *self.blocks if isinstance(self, ChainBlock) else (self,),
                *other.blocks if isinstance(other, ChainBlock) else (other,),
            )
        else:
            if isinstance(self, ScaleBlock):
                p = self.parameters.parameter
                return ScaleBlock(self.block, p * other)
            else:
                return ScaleBlock(self, other if isinstance(other, Parameter) else Parameter(other))

    def __truediv__(self, other: Union[TNumber, sympy.Basic]) -> AbstractBlock:
        if not isinstance(other, (get_args(TNumber), sympy.Basic)):
            raise TypeError("Cannot divide block by another block.")
        ix = 1 / other
        return self * ix  # type: ignore [no-any-return]

    def __add__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.utils import add

        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only add a block to another block. Got {type(other)}.")
        return add(self, other)

    def __radd__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.utils import add

        if isinstance(other, int) and other == 0:
            return self
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only add a block to another block. Got {type(other)}.")
        return add(other, self)

    def __iadd__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.composite import AddBlock
        from qadence.blocks.utils import add

        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only add a block to another block. Got {type(other)}.")

        # We make sure to unroll any AddBlocks, because for iadd we
        # assume the user expected in-place addition
        return add(
            *self.blocks if isinstance(self, AddBlock) else (self,),
            *other.blocks if isinstance(other, AddBlock) else (other,),
        )

    def __sub__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.primitive import ScaleBlock
        from qadence.blocks.utils import add

        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only subtract a block from another block. Got {type(other)}.")
        if isinstance(other, ScaleBlock):
            scale = other.parameters.parameter
            b = ScaleBlock(other.block, -scale)
        else:
            b = ScaleBlock(other, Parameter(-1.0))
        return add(self, b)

    def __isub__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.composite import AddBlock
        from qadence.blocks.utils import add

        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only add a block to another block. Got {type(other)}.")

        # We make sure to unroll (and minus) any AddBlocks: for isub we assume the
        # user expected in-place subtraction
        return add(
            *self.blocks if isinstance(self, AddBlock) else (self,),
            *(-block for block in other.blocks) if isinstance(other, AddBlock) else (-other,),
        )

    def __pow__(self, power: int) -> AbstractBlock:
        from qadence.blocks.utils import chain

        return chain(self for _ in range(power))

    def __neg__(self) -> AbstractBlock:
        return self.__mul__(-1.0)

    def __pos__(self) -> AbstractBlock:
        return self

    def __matmul__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.utils import kron

        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only kron a block to another block. Got {type(other)}.")
        return kron(self, other)

    def __imatmul__(self, other: AbstractBlock) -> AbstractBlock:
        from qadence.blocks.composite import KronBlock
        from qadence.blocks.utils import kron

        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Can only kron a block with another block. Got {type(other)}.")

        # We make sure to unroll any KronBlocks, because for ixor we assume the user
        # expected in-place kron
        return kron(
            *self.blocks if isinstance(self, KronBlock) else (self,),
            *other.blocks if isinstance(other, KronBlock) else (other,),
        )

    def __iter__(self) -> Iterable:
        yield self

    def __len__(self) -> int:
        return 1

    @property
    def _block_title(self) -> str:
        bits = ",".join(str(i) for i in self.qubit_support)
        s = f"{type(self).__name__}({bits})"

        if self.tag is not None:
            s += rf" \[tag: {self.tag}]"
        return s

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            return Tree(self._block_title)
        else:
            tree.add(self._block_title)
        return tree

    def __repr__(self) -> str:
        console = Console()
        with console.capture() as cap:
            console.print(self.__rich_tree__())
        return cap.get().strip()  # type: ignore [no-any-return]

    @abstractproperty
    def depth(self) -> int:
        pass

    @abstractmethod
    def __ascii__(self, console: Console) -> RenderableType:
        pass

    @abstractmethod
    def _to_dict(self) -> dict:
        pass

    @classmethod
    @abstractmethod
    def _from_dict(cls, d: dict) -> AbstractBlock:
        pass

    def _to_json(self) -> str:
        return json.dumps(self._to_dict())

    @classmethod
    def _from_json(cls, path: str | Path) -> AbstractBlock:
        d: dict = {}
        if isinstance(path, str):
            path = Path(path)
        try:
            with open(path, "r") as file:
                d = json.load(file)

        except Exception as e:
            print(f"Unable to load block due to {e}")

        return AbstractBlock._from_dict(d)

    def _to_file(self, path: str | Path = Path("")) -> None:
        if isinstance(path, str):
            path = Path(path)
        try:
            with open(path, "w") as file:
                file.write(self._to_json())
        except Exception as e:
            print(f"Unable to write {type(self)} to disk due to {e}")

    def __hash__(self) -> int:
        return hash(self._to_json())

    @abstractmethod
    def dagger(self) -> AbstractBlock:
        raise NotImplementedError(
            f"Hermitian adjoint of the Block '{type(self)}' is not implemented yet!"
        )

    @property
    def is_parametric(self) -> bool:
        from qadence.blocks.utils import parameters

        params: list[sympy.Basic] = parameters(self)
        return any(isinstance(p, Parameter) for p in params)

    def tensor(self, values: dict[str, TNumber | torch.Tensor] = {}) -> torch.Tensor:
        from .block_to_tensor import block_to_tensor

        return block_to_tensor(self, values)

    @property
    def _is_diag_pauli(self) -> bool:
        from qadence.blocks import CompositeBlock, PrimitiveBlock, ScaleBlock
        from qadence.blocks.utils import block_is_qubit_hamiltonian

        if not block_is_qubit_hamiltonian(self):
            return False

        elif isinstance(self, CompositeBlock):
            return all([b._is_diag_pauli for b in self.blocks])

        elif isinstance(self, ScaleBlock):
            return self.block._is_diag_pauli
        elif isinstance(self, PrimitiveBlock):
            return self.name in ["Z", "I"]
        return False

    @property
    def is_identity(self) -> bool:
        """Identity predicate for blocks."""
        from qadence.blocks import CompositeBlock, PrimitiveBlock, ScaleBlock

        if isinstance(self, CompositeBlock):
            return all([b.is_identity for b in self.blocks])
        elif isinstance(self, ScaleBlock):
            return self.block.is_identity
        elif isinstance(self, PrimitiveBlock):
            return self.name == "I"
        return False


TAbstractBlock = TypeVar("TAbstractBlock", bound=AbstractBlock)



from __future__ import annotations

from collections import Counter
from math import log2
from typing import Callable, Sequence

import numpy as np
import pyqtorch as pyq
import torch
from pyqtorch.apply import apply_operator
from pyqtorch.parametric import Parametric as PyQParametric
from torch import (
    Tensor,
    cat,
    complex128,
    mean,
    no_grad,
    rand,
)

from qadence.types import ParamDictType
from qadence.utils import Endianness, int_to_basis, is_qadence_shape

FINITE_DIFF_EPS = 1e-06
# Dict of NumPy dtype -> torch dtype (when the correspondence exists)
numpy_to_torch_dtype_dict = {
    np.bool_: torch.bool,
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
    int: torch.int64,
    float: torch.float64,
    complex: torch.complex128,
}


def param_dict(keys: Sequence[str], values: Sequence[Tensor]) -> dict[str, Tensor]:
    return {key: val for key, val in zip(keys, values)}


def numpy_to_tensor(
    x: np.ndarray,
    device: torch.device = torch.device("cpu"),
    dtype: torch.dtype = torch.complex128,
    requires_grad: bool = False,
) -> Tensor:
    """This only copies the numpy array if device or dtype are different than the ones of x."""
    return torch.as_tensor(x, dtype=dtype, device=device).requires_grad_(requires_grad)


def promote_to_tensor(
    x: Tensor | np.ndarray | float,
    dtype: torch.dtype = torch.complex128,
    requires_grad: bool = True,
) -> Tensor:
    """Convert the given type into a torch.Tensor."""
    if isinstance(x, float):
        return torch.tensor([[x]], dtype=dtype, requires_grad=requires_grad)
    elif isinstance(x, np.ndarray):
        return numpy_to_tensor(
            x, dtype=numpy_to_torch_dtype_dict.get(x.dtype), requires_grad=requires_grad
        )
    elif isinstance(x, Tensor):
        return x.requires_grad_(requires_grad)
    else:
        raise ValueError(f"Don't know how to promote {type(x)} to Tensor")


# FIXME: Not being used, maybe remove in v1.0.0
def count_bitstrings(sample: Tensor, endianness: Endianness = Endianness.BIG) -> Counter:
    # Convert to a tensor of integers.
    n_qubits = sample.size()[1]
    base = torch.ones(n_qubits, dtype=torch.int64) * 2
    powers_of_2 = torch.pow(base, reversed(torch.arange(n_qubits)))
    int_tensor = torch.matmul(sample, powers_of_2)
    # Count occurences of integers.
    count_int = torch.bincount(int_tensor)
    # Return a Counter for non-empty bitstring counts.
    return Counter(
        {
            int_to_basis(k=k, n_qubits=n_qubits, endianness=endianness): count.item()
            for k, count in enumerate(count_int)
            if count > 0
        }
    )


def to_list_of_dicts(param_values: ParamDictType) -> list[ParamDictType]:
    if not param_values:
        return [param_values]

    max_batch_size = max(p.size()[0] for p in param_values.values())
    batched_values = {
        k: (v if v.size()[0] == max_batch_size else v.repeat(max_batch_size, 1))
        for k, v in param_values.items()
    }

    return [{k: v[i] for k, v in batched_values.items()} for i in range(max_batch_size)]


def pyqify(state: Tensor, n_qubits: int = None) -> Tensor:
    """Convert a state of shape (batch_size, 2**n_qubits) to [2] * n_qubits + [batch_size]."""
    if n_qubits is None:
        n_qubits = int(log2(state.shape[1]))
    if (state.ndim != 2) or (state.size(1) != 2**n_qubits):
        raise ValueError(
            "The initial state must be composed of tensors of size "
            f"(batch_size, 2**n_qubits). Found: {state.size() = }."
        )
    return state.T.reshape([2] * n_qubits + [state.size(0)])


def unpyqify(state: Tensor) -> Tensor:
    """Convert a state of shape [2] * n_qubits + [batch_size] to (batch_size, 2**n_qubits)."""
    return torch.flatten(state, start_dim=0, end_dim=-2).t()


def is_pyq_shape(state: Tensor, n_qubits: int) -> bool:
    return state.size()[:-1] == [2] * n_qubits  # type: ignore[no-any-return]


def validate_state(state: Tensor, n_qubits: int) -> None:
    """Check if a custom initial state conforms to the qadence or the pyqtorch format."""
    if state.dtype != complex128:
        raise TypeError(f"Expected type complex128, got {state.dtype}")
    elif len(state.size()) < 2:
        raise ValueError(f"Invalid state shape. Got {state.shape}")
    elif not is_qadence_shape(state, n_qubits) and not is_pyq_shape(state, n_qubits):
        raise ValueError(
            f"Allowed formats for custom initial state are:\
                  (1) Qadence shape: (batch_size, 2**n_qubits)\
                  (2) Pyqtorch shape: (2 * n_qubits + [batch_size])\
                  Found: {state.size() = }"
        )


def infer_batchsize(param_values: ParamDictType = None) -> int:
    """Infer the batch_size through the length of the parameter tensors."""
    return max([len(tensor) for tensor in param_values.values()]) if param_values else 1


# The following functions can be used to compute potentially higher order gradients using pyqtorch's
# native 'jacobian' methods.


def finitediff(f: Callable, x: Tensor, eps: float = FINITE_DIFF_EPS) -> Tensor:
    return (f(x + eps) - f(x - eps)) / (2 * eps)  # type: ignore


def finitediff_sampling(
    f: Callable, x: Tensor, eps: float = FINITE_DIFF_EPS, num_samples: int = 10
) -> Tensor:
    def _finitediff(val: Tensor) -> Tensor:
        return (f(x + val) - f(x - val)) / (2 * val)  # type: ignore

    with no_grad():
        return mean(cat([_finitediff(val) for val in rand(1) for _ in range(num_samples)]))


def dydx(
    jacobian: Tensor,
    qubit_support: tuple,
    out_state: Tensor,
    projected_state: Tensor,
) -> Tensor:
    return 2 * pyq.overlap(
        projected_state,
        apply_operator(
            state=out_state,
            operator=jacobian,
            qubits=qubit_support,
        ),
    )


def dydxx(
    op: PyQParametric,
    values: dict[str, Tensor],
    out_state: Tensor,
    projected_state: Tensor,
) -> Tensor:
    return 2 * finitediff_sampling(
        lambda val: dydx(
            op.jacobian({op.param_name: val}), op.qubit_support, out_state, projected_state
        ),
        values[op.param_name],
    )



from __future__ import annotations

from typing import Any

from pyqtorch.apply import apply_operator
from pyqtorch.circuit import QuantumCircuit as PyQCircuit
from pyqtorch.parametric import Parametric as PyQParametric
from pyqtorch.primitive import Primitive as PyQPrimitive
from pyqtorch.utils import overlap, param_dict
from torch import Tensor, no_grad, tensor
from torch.autograd import Function
from torch.nn import Module

from qadence.backends.pyqtorch.convert_ops import PyQHamiltonianEvolution, ScalePyQOperation
from qadence.blocks.abstract import AbstractBlock


class AdjointExpectation(Function):
    """
    The adjoint differentiation method (https://arxiv.org/pdf/2009.02823.pdf).

    is able to perform a backward pass in O(P) time and maintaining
    atmost 3 states where P is the number of parameters in a variational circuit.

    Pseudo-code of the algorithm:

    c: a variational circuit
    c.gates = gate0,gate1,..gateN, where N denotes the last gate in c
    o: a observable
    state: an initial state

    1. Forward pass.
    for gate in c.gates: # We apply gate0, gate1 to gateN.
        state = gate(state)

    projected_state = o(state)  # Apply the observable to the state.
    expval = overlap(state, projected_state)  # Compute the expected value.

    2. Backward pass:
    grads = []
    for gate in reversed(c.gates): # Iterate through c.gates in "reverse", so gateN, gateN-1 etc.
        state = dagger(gate)(state)  # 'Undo' the gate by applying its dagger.
        if gate is Parametric:
            mu = jacobian(gate)(state) # Compute the jacobian of the gate w.r.t its parameter.
            grads.append(2 * overlap(mu, projected_state) # Compute the gradient.
        projected_state = dagger(gate)(projected_state)  # 'Undo' the gate from the projected_state.

    Current Limitations:

    (1) The adjoint method is only available in the pyqtorch backend.
    (2) Parametric observables are not supported.
    (3) Multiple observables are not supported.
    (4) Higher order derivatives are not natively supported.
    (5) Only expectation values can be differentiated, not wave functions.
    """

    @staticmethod
    @no_grad()
    def forward(
        ctx: Any,
        circuit: PyQCircuit,
        observable: PyQCircuit,
        state: Tensor,
        param_names: list[str],
        *param_values: Tensor,
    ) -> Tensor:
        for param in param_values:
            param = param.detach()
        ctx.circuit = circuit
        ctx.observable = observable
        ctx.param_names = param_names
        values = param_dict(param_names, param_values)
        ctx.out_state = circuit.run(state, values)
        ctx.projected_state = observable.run(ctx.out_state, values)
        ctx.save_for_backward(*param_values)
        return overlap(ctx.out_state, ctx.projected_state)

    @staticmethod
    @no_grad()
    def backward(ctx: Any, grad_out: Tensor) -> tuple:
        param_values = ctx.saved_tensors
        values = param_dict(ctx.param_names, param_values)

        def _apply_adjoint(ctx: Any, op: Module) -> list:
            grads: list = []
            if isinstance(op, PyQHamiltonianEvolution):
                generator = op.block.generator
                time_param = values[op.param_names[0]]
                ctx.out_state = apply_operator(ctx.out_state, op.dagger(values), op.qubit_support)
                # A HamEvo can have a parametrized (1) time evolution and/or (2) generator.
                if (
                    isinstance(generator, AbstractBlock)
                    and generator.is_parametric
                    and values[op.param_names[1]].requires_grad
                ):
                    # If the generator contains a trainable parameter, we compute its gradient.
                    mu = apply_operator(
                        ctx.out_state, op.jacobian_generator(values), op.qubit_support
                    )
                    grads.append(2 * overlap(ctx.projected_state, mu))
                if time_param.requires_grad:
                    # If the time evolution is trainable, we compute its gradient.
                    mu = apply_operator(ctx.out_state, op.jacobian_time(values), op.qubit_support)
                    grads.append(2 * overlap(ctx.projected_state, mu))
                ctx.projected_state = apply_operator(
                    ctx.projected_state, op.dagger(values), op.qubit_support
                )
            elif isinstance(op, ScalePyQOperation):
                ctx.out_state = apply_operator(ctx.out_state, op.dagger(values), op.qubit_support)
                scaled_pyq_op = op.operations[0]
                if (
                    isinstance(scaled_pyq_op, PyQParametric)
                    and values[scaled_pyq_op.param_name].requires_grad
                ):
                    mu = apply_operator(
                        ctx.out_state,
                        scaled_pyq_op.jacobian(values),
                        scaled_pyq_op.qubit_support,
                    )
                    grads.append(2 * overlap(ctx.projected_state, mu))

                if values[op.param_name].requires_grad:
                    grads.append(2 * -values[op.param_name])
                ctx.projected_state = apply_operator(
                    ctx.projected_state, op.dagger(values), op.qubit_support
                )
            elif isinstance(op, PyQCircuit):
                grads = [g for sub_op in op.reverse() for g in _apply_adjoint(ctx, sub_op)]
            elif isinstance(op, PyQPrimitive):
                ctx.out_state = apply_operator(ctx.out_state, op.dagger(values), op.qubit_support)
                if isinstance(op, PyQParametric) and values[op.param_name].requires_grad:
                    mu = apply_operator(
                        ctx.out_state,
                        op.jacobian(values),
                        op.qubit_support,
                    )
                    grads.append(2 * overlap(ctx.projected_state, mu))
                ctx.projected_state = apply_operator(
                    ctx.projected_state, op.dagger(values), op.qubit_support
                )
            else:
                raise TypeError(
                    f"AdjointExpectation does not support a backward pass for type {type(op)}."
                )

            return grads

        grads = list(
            reversed(
                [grad_out * g for op in ctx.circuit.reverse() for g in _apply_adjoint(ctx, op)]
            )
        )
        num_grads = len(grads)
        num_params = len(ctx.saved_tensors)
        diff = num_params - num_grads
        grads = grads + [tensor([0]) for _ in range(diff)]
        # Set observable grads to 0
        ctx.save_for_backward(*grads)
        return (None, None, None, None, *grads)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Tuple

import jax.numpy as jnp
from jax import Array, custom_vjp, vmap

from qadence.backend import Backend as QuantumBackend
from qadence.backend import ConvertedCircuit, ConvertedObservable
from qadence.backends.jax_utils import (
    tensor_to_jnp,
)
from qadence.blocks.utils import uuid_to_eigen
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.noise import Noise
from qadence.types import Endianness, Engine, ParamDictType


def compute_single_gap(eigen_vals: Array, default_val: float = 2.0) -> Array:
    eigen_vals = eigen_vals.reshape(1, 2)
    gaps = jnp.abs(jnp.tril(eigen_vals.T - eigen_vals))
    return jnp.unique(jnp.where(gaps > 0.0, gaps, default_val), size=1)


@dataclass
class DifferentiableExpectation:
    """A handler for differentiating expectation estimation using various engines."""

    backend: QuantumBackend
    circuit: ConvertedCircuit
    observable: list[ConvertedObservable] | ConvertedObservable
    param_values: ParamDictType
    state: Array | None = None
    measurement: Measurements | None = None
    noise: Noise | None = None
    mitigation: Mitigations | None = None
    endianness: Endianness = Endianness.BIG
    engine: Engine = Engine.JAX

    def psr(self) -> Any:
        n_obs = len(self.observable)

        def expectation_fn(state: Array, values: ParamDictType, psr_params: ParamDictType) -> Array:
            return self.backend.expectation(
                circuit=self.circuit, observable=self.observable, param_values=values, state=state
            )

        @custom_vjp
        def expectation(state: Array, values: ParamDictType, psr_params: ParamDictType) -> Array:
            return expectation_fn(state, values, psr_params)

        uuid_to_eigs = {
            k: tensor_to_jnp(v) for k, v in uuid_to_eigen(self.circuit.abstract.block).items()
        }
        self.psr_params = {
            k: self.param_values[k] for k in uuid_to_eigs.keys()
        }  # Subset of params on which to perform PSR.

        def expectation_fwd(state: Array, values: ParamDictType, psr_params: ParamDictType) -> Any:
            return expectation_fn(state, values, psr_params), (
                state,
                values,
                psr_params,
            )

        def expectation_bwd(res: Tuple[Array, ParamDictType, ParamDictType], tangent: Array) -> Any:
            state, values, psr_params = res
            grads = {}
            # FIXME Hardcoding the single spectral_gap to 2.
            spectral_gap = 2.0
            shift = jnp.pi / 2

            def shift_circ(param_name: str, values: dict) -> Array:
                shifted_values = values.copy()
                shiftvals = jnp.array(
                    [shifted_values[param_name] + shift, shifted_values[param_name] - shift]
                )

                def _expectation(val: Array) -> Array:
                    shifted_values[param_name] = val
                    return expectation(state, shifted_values, psr_params)

                return vmap(_expectation, in_axes=(0,))(shiftvals)

            for param_name, _ in psr_params.items():
                f_plus, f_min = shift_circ(param_name, values)
                grad = spectral_gap * (f_plus - f_min) / (4.0 * jnp.sin(spectral_gap * shift / 2.0))
                grads[param_name] = jnp.sum(tangent * grad, axis=1) if n_obs > 1 else tangent * grad
            return None, None, grads

        expectation.defvjp(expectation_fwd, expectation_bwd)
        return expectation(self.state, self.param_values, self.psr_params)



from __future__ import annotations

from abc import abstractmethod
from copy import deepcopy
from typing import Any, Iterable, Tuple

import sympy
import torch
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.tree import Tree

from qadence.blocks.abstract import AbstractBlock
from qadence.parameters import (
    Parameter,
    ParamMap,
    dagger_expression,
    evaluate,
    extract_original_param_entry,
    stringify,
)
from qadence.types import TParameter
from qadence.utils import format_parameter


class PrimitiveBlock(AbstractBlock):
    """
    Primitive blocks represent elementary unitary operations.

    Examples are single/multi-qubit gates or Hamiltonian evolution.
    See [`qadence.operations`](/qadence/operations.md) for a full list of
    primitive blocks.
    """

    name = "PrimitiveBlock"

    def __init__(self, qubit_support: tuple[int, ...]):
        self._qubit_support = qubit_support

    @property
    def qubit_support(self) -> Tuple[int, ...]:
        return self._qubit_support

    def digital_decomposition(self) -> AbstractBlock:
        """Decomposition into purely digital gates.

        This method returns a decomposition of the Block in a
        combination of purely digital single-qubit and two-qubit
        'gates', by manual/custom knowledge of how this can be done efficiently.
        :return:
        """
        return self

    def __len__(self) -> int:
        return 1

    def __iter__(self) -> Iterable:
        yield self

    @property
    def depth(self) -> int:
        return 1

    def __ascii__(self, console: Console) -> RenderableType:
        return Panel(self._block_title, expand=False)

    def __xor__(self, other: int | AbstractBlock) -> AbstractBlock:
        if isinstance(other, int):
            from qadence.transpile import repeat

            B = type(self)
            (start,) = self.qubit_support
            return repeat(B, range(start, start + other))
        else:
            raise TypeError(f"PrimitiveBlocks cannot use ^ on type {type(other)}")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, type(self)):
            return self.qubit_support == other.qubit_support
        return False

    def _to_dict(self) -> dict:
        return {
            "type": type(self).__name__,
            "qubit_support": self.qubit_support,
            "tag": self.tag,
        }

    @classmethod
    def _from_dict(cls, d: dict) -> PrimitiveBlock:
        return cls(*d["qubit_support"])

    def __hash__(self) -> int:
        return hash(self._to_json())

    @property
    def n_qubits(self) -> int:
        return max(self.qubit_support) + 1

    @property
    def n_supports(self) -> int:
        return len(self.qubit_support)

    def dagger(self) -> PrimitiveBlock:
        return self


class ParametricBlock(PrimitiveBlock):
    """Parameterized primitive blocks."""

    name = "ParametricBlock"

    # a tuple of Parameter's specifies which parameters go into this block
    parameters: ParamMap

    # any unitary can be written as exp(iH).
    # For a parametric block this is particularly interesting and
    # is known for most basic 'gates' or analog pulses.
    generator: AbstractBlock | Parameter | TParameter | None = None

    @property
    def _block_title(self) -> str:
        s = super()._block_title
        params_str = []
        for p in self.parameters.expressions():
            if p.is_number:
                val = evaluate(p)
                if isinstance(val, float):
                    val = round(val, 2)
                params_str.append(val)
            else:
                params_str.append(stringify(p))

        return s + rf" \[params: {params_str}]"

    @property
    def trainable(self) -> bool:
        for expr in self.parameters.expressions():
            if expr.is_number:
                return False
            else:
                return any(not p.trainable for p in expr.free_symbols)
        return True

    @abstractmethod
    def num_parameters(cls) -> int:
        """The number of parameters required by the block.

        This is a class property since the number of parameters is defined
        automatically before instantiating the operation. Also, this could
        correspond to a larger number of actual user-facing parameters
        since any parameter expression is allowed

        Examples:
        - RX operation has 1 parameter
        - U operation has 3 parameters
        - HamEvo has 2 parameters (generator and time evolution)
        """
        pass

    def __xor__(self, other: int | AbstractBlock) -> AbstractBlock:
        if isinstance(other, AbstractBlock):
            return super().__xor__(other)
        elif isinstance(other, int):
            from qadence.transpile import repeat

            B = type(self)
            (start,) = self.qubit_support
            (param,) = self.parameters.expressions()
            return repeat(B, range(start, start + other), stringify(param))
        else:
            raise ValueError(f"PrimitiveBlocks cannot use ^ on type {type(other)}")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, type(self)):
            return (
                self.qubit_support == other.qubit_support
                and self.parameters.parameter == other.parameters.parameter
            )
        return False

    def __contains__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            raise TypeError(f"Cant check if {type(other)} in {type(self)}")
        for p in self.parameters.expressions():
            if other in p.free_symbols:
                return True
        return False

    def _to_dict(self) -> dict:
        return {
            "type": type(self).__name__,
            "qubit_support": self.qubit_support,
            "tag": self.tag,
            "parameters": self.parameters._to_dict(),
        }

    @classmethod
    def _from_dict(cls, d: dict) -> ParametricBlock:
        params = ParamMap._from_dict(d["parameters"])
        target = d["qubit_support"][0]
        return cls(target, params)  # type: ignore[call-arg]

    def dagger(self) -> ParametricBlock:
        exprs = self.parameters.expressions()
        params = tuple(-extract_original_param_entry(param) for param in exprs)
        return type(self)(*self.qubit_support, *params)  # type: ignore[arg-type]


class ScaleBlock(ParametricBlock):
    """Scale blocks are created when multiplying a block by a number or parameter.

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence import X

    print(X(0) * 2)
    ```
    """

    name = "ScaleBlock"

    block: AbstractBlock

    def __init__(self, block: AbstractBlock, parameter: Any):
        self.block = block
        # TODO: more meaningful name like `scale`?
        self.parameters = (
            parameter if isinstance(parameter, ParamMap) else ParamMap(parameter=parameter)
        )
        super().__init__(block.qubit_support)

    def __pow__(self, power: int) -> AbstractBlock:
        from qadence.blocks.utils import chain

        expr = self.parameters.parameter
        return ScaleBlock(chain(self.block for _ in range(power)), expr**power)

    @property
    def qubit_support(self) -> Tuple[int, ...]:
        return self.block.qubit_support

    @classmethod
    def num_parameters(cls) -> int:
        return 1

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        return self.block.eigenvalues_generator

    @property
    def eigenvalues(self) -> torch.Tensor:
        return self.block.eigenvalues

    @property
    def _block_title(self) -> str:
        (scale,) = self.parameters.expressions()
        s = rf"\[mul: {format_parameter(scale)}] "
        return s

    @property
    def n_qubits(self) -> int:
        return self.block.n_qubits

    @property
    def scale(self) -> sympy.Expr:
        (scale,) = self.parameters.expressions()
        return scale

    def __rich_tree__(self, tree: Tree = None) -> Tree:
        if tree is None:
            tree = Tree(self._block_title)
        else:
            tree = tree.add(self._block_title)
        self.block.__rich_tree__(tree)
        return tree

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        elif isinstance(other, ScaleBlock):
            return (
                self.block == other.block
                and self.parameters.parameter == other.parameters.parameter
            )
        return False

    def __contains__(self, other: object) -> bool:
        from qadence.blocks.composite import CompositeBlock

        if isinstance(other, AbstractBlock):
            if isinstance(self.block, CompositeBlock) and other in self.block:
                return True
            else:
                return self.block == other

        if isinstance(other, Parameter):
            if isinstance(self.block, ParametricBlock) or isinstance(self.block, CompositeBlock):
                return other in self.block
            return False
        else:
            raise TypeError(
                f"Can not check for containment between {type(self)} and {type(other)}."
            )

    def dagger(self) -> ScaleBlock:
        p = list(self.parameters.expressions())[0]
        return self.__class__(self.block.dagger(), dagger_expression(p))

    def _to_dict(self) -> dict:
        return {
            "type": type(self).__name__,
            "tag": self.tag,
            "parameters": self.parameters._to_dict(),
            "block": self.block._to_dict(),
        }

    @classmethod
    def _from_dict(cls, d: dict) -> ScaleBlock:
        from qadence import blocks as qadenceblocks
        from qadence import operations

        expr = ParamMap._from_dict(d["parameters"])
        block: AbstractBlock
        if hasattr(operations, d["block"]["type"]):
            block = getattr(operations, d["block"]["type"])._from_dict(d["block"])

        else:
            block = getattr(qadenceblocks, d["block"]["type"])._from_dict(d["block"])
        return cls(block, expr)  # type: ignore[arg-type]


class TimeEvolutionBlock(ParametricBlock):
    """Simple time evolution block with time-independent Hamiltonian.

    This class is just a convenience class which is used to label
    blocks which contains simple time evolution with time-independent
    Hamiltonian operators
    """

    name = "TimeEvolutionBlock"

    @property
    def has_parametrized_generator(self) -> bool:
        return not isinstance(self.generator, AbstractBlock)


class ControlBlock(PrimitiveBlock):
    """The abstract ControlBlock."""

    name = "Control"
    control: tuple[int, ...]
    target: tuple[int, ...]

    def __init__(self, control: tuple[int, ...], target_block: PrimitiveBlock) -> None:
        self.control = control
        self.blocks = (target_block,)
        self.target = target_block.qubit_support

        # using tuple expansion because some control operations could
        # have multiple targets, e.g. CSWAP
        super().__init__((*control, *self.target))  # target_block.qubit_support[0]))

    @property
    def n_controls(self) -> int:
        return len(self.control)

    @property
    def n_targets(self) -> int:
        return len(self.target)

    @property
    def _block_title(self) -> str:
        s = f"{self.name}{self.qubit_support}"
        return s if self.tag is None else (s + rf" \[tag: {self.tag}]")

    def __ascii__(self, console: Console) -> RenderableType:
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, type(self)):
            return self.qubit_support == other.qubit_support and self.blocks[0] == other.blocks[0]
        return False

    def _to_dict(self) -> dict:
        return {
            "type": type(self).__name__,
            "qubit_support": self.qubit_support,
            "tag": self.tag,
            "blocks": [b._to_dict() for b in self.blocks],
        }

    @classmethod
    def _from_dict(cls, d: dict) -> ControlBlock:
        control = d["qubit_support"][0]
        target = d["qubit_support"][1]
        return cls(control, target)

    def dagger(self) -> ControlBlock:
        blk = deepcopy(self)
        blk.blocks = (self.blocks[0].dagger(),)
        return blk


class ParametricControlBlock(ParametricBlock):
    """The abstract parametrized ControlBlock."""

    name = "ParameterizedControl"
    control: tuple[int, ...] = ()
    blocks: tuple[ParametricBlock, ...]

    def __init__(self, control: tuple[int, ...], target_block: ParametricBlock) -> None:
        self.blocks = (target_block,)
        self.control = control
        self.parameters = target_block.parameters
        super().__init__((*control, *target_block.qubit_support))

    @property
    def n_controls(self) -> int:
        return len(self.control)

    @property
    def eigenvalues_generator(self) -> torch.Tensor:
        return torch.empty(0)

    def __ascii__(self, console: Console) -> RenderableType:
        raise NotImplementedError

    @property
    def n_qubits(self) -> int:
        return len(self.qubit_support)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractBlock):
            raise TypeError(f"Cant compare {type(self)} to {type(other)}")
        if isinstance(other, type(self)):
            return self.qubit_support == other.qubit_support and self.blocks[0] == other.blocks[0]
        return False

    def _to_dict(self) -> dict:
        return {
            "type": type(self).__name__,
            "qubit_support": self.qubit_support,
            "tag": self.tag,
            "blocks": [b._to_dict() for b in self.blocks],
        }

    @classmethod
    def _from_dict(cls, d: dict) -> ParametricControlBlock:
        from qadence.serialization import deserialize

        control = d["qubit_support"][0]
        target = d["qubit_support"][1]
        targetblock = d["blocks"][0]
        expr = deserialize(targetblock["parameters"])
        block = cls(control, target, expr)  # type: ignore[call-arg]
        return block

    @property
    def _block_title(self) -> str:
        s = f"{self.name}{self.qubit_support}"
        params_str = []
        for p in self.parameters.expressions():
            if p.is_number:
                val = evaluate(p)
                if isinstance(val, float):
                    val = round(val, 2)
                params_str.append(val)
            else:
                params_str.append(stringify(p))

        s += rf" \[params: {params_str}]"
        return s if self.tag is None else (s + rf" \[tag: {self.tag}]")

    def dagger(self) -> ParametricControlBlock:
        blk = deepcopy(self)
        blocks = tuple(b.dagger() for b in blk.blocks)
        blk.blocks = blocks
        blk.parameters = blocks[0].parameters
        return blk


class ProjectorBlock(PrimitiveBlock):
    """The abstract ProjectorBlock."""

    name = "ProjectorBlock"

    def __init__(
        self,
        ket: str,
        bra: str,
        qubit_support: int | tuple[int, ...],
    ) -> None:
        """
        Arguments:

            ket (str): The ket given as a bitstring.
            bra (str): The bra given as a bitstring.
            qubit_support (int | tuple[int]): The qubit_support of the block.
        """
        if isinstance(qubit_support, int):
            qubit_support = (qubit_support,)
        if len(bra) != len(ket):
            raise ValueError(
                "Bra and ket must be bitstrings of same length in the 'Projector' definition."
            )
        elif len(bra) != len(qubit_support):
            raise ValueError("Bra or ket must be of same length as the 'qubit_support'")
        for wf in [bra, ket]:
            if not all(int(item) == 0 or int(item) == 1 for item in wf):
                raise ValueError(
                    "All qubits must be either in the '0' or '1' state"
                    " in the 'ProjectorBlock' definition."
                )

        self.ket = ket
        self.bra = bra
        super().__init__(qubit_support)
